#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Reads SLA file and use Pillow to draw a schematic picture of all page
and page objects.
"""

# Imports ===============================================================#

import math

from PIL import Image, ImageDraw

import pyscribus.dimensions as dimensions
import pyscribus.pageobjects as pageobjects
import pyscribus.pages as pages

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

class WireframeObject:
    """
    Wireframe object : something to draw on the wireframe, like pages,
    page objects.

    :type sla_object: pyscribus.pages.Page, pyscribus.pageobjects.PageObject
    :param sla_object: PyScribus instance of drawnable object
    """

    def __init__(self, sla_object=False):
        self.box = False
        self.bleed = False

        self.is_page = False
        self.type = "undefined"
        self.group_objects = []

        # Pages have layer -1, as 0 is the lower layer available
        self.layer = -1

        self.draw_settings = {
            "fill": "",
            "outline": "",
            "bleed": "red"
        }

        if sla_object:
            self.from_object(sla_object)

    def from_object(self, sla_object):

        if isinstance(sla_object, pages.Page):
            self.is_page = True
            self.type = "page"
        else:
            self.is_page = False
            type_ok = False

            for class_test in [
                    pageobjects.LatexObject,
                    pageobjects.RenderObject]:

                if isinstance(sla_object, class_test):
                    self.type = "render"
                    type_ok = True
                    break

            if not type_ok:

                for human, class_test in pageobjects.po_type_classes.items():

                    if isinstance(sla_object, class_test):
                        self.type = human
                        type_ok = True
                        break

        if self.type == "group":
            self.group_objects = sla_object.group_objects

        if isinstance(sla_object, pageobjects.PageObject):

            if sla_object.rotated:
                self.box = sla_object.rotated_box
            else:
                self.box = sla_object.box

            self.layer = sla_object.layer

        else:
            self.box = sla_object.box

    def draw_on_canvas(self, canvas, bleed=False):

        if bleed and self.bleed:
            tx = self.box.coords["top-left"][0].value - self.bleed["left"]
            ty = self.box.coords["top-left"][1].value - self.bleed["top"]
            bx = self.box.coords["bottom-right"][0].value + self.bleed["right"]
            by = self.box.coords["bottom-right"][1].value + self.bleed["bottom"]

            bleed_rect = ((tx, ty), (bx, by))
            canvas.rectangle(bleed_rect, outline="red")

        rect = (
            (
                self.box.coords["top-left"][0],
                self.box.coords["top-left"][1]
            ),
            (
                self.box.coords["bottom-right"][0],
                self.box.coords["bottom-right"][1]
            ),
        )

        if self.draw_settings["fill"] and self.draw_settings["outline"]:

            canvas.rectangle(
                rect,
                fill=self.draw_settings["fill"],
                outline=self.draw_settings["outline"]
            )

        else:

            if self.draw_settings["fill"]:
                canvas.rectangle(rect, fill=self.draw_settings["fill"])

            if self.draw_settings["outline"]:
                canvas.rectangle(rect, outline=self.draw_settings["outline"])

        if self.type == "group":

            for subpo in self.group_objects:
                s = WireframeObject(subpo)
                s.draw_on_canvas(canvas, bleed)

        return canvas


class Wireframe:
    """
    Wireframe canvas.

    Reads SLA file and use Pillow to draw a schematic picture of all page
    and page objects.

    :ivar list pages: List of pages as WireframeObject
    :ivar list page_objects: List of page objects as WireframeObject
    """

    stylesheet = {
        "page": {"fill": "white", "outline": ""},
        "text": {"fill": "lightgrey", "outline": "black"},
        "line": {"fill": "", "outline": "black"},
        "image": {"fill": "lightblue", "outline": "blue"},
        "table": {"fill": "", "outline": "darkgreen"},
        "render": {"fill": "lightgreen", "outline": "black"},
        "polygon": {"fill": "yellow", "outline": "black"},
    }

    def __init__(self):
        self.pages = []
        self.page_objects = []

        self.bleed = {
            "top": dimensions.Dim(0), "right": dimensions.Dim(0),
            "left": dimensions.Dim(0), "bottom": dimensions.Dim(0)
        }

    def from_sla(self, sla):
        """
        Load all drawnable objects of sla file.

        :type sla: string
        :param sla: SLA file path
        """
        self.bleed = sla.document.bleed

        for page in sla.document.pages:
            wo = WireframeObject(page)
            wo.bleed = self.bleed
            self.pages.append(wo)

        for pago in sla.document.page_objects:
            wo = WireframeObject(pago)
            self.page_objects.append(wo)

    def append(self, sla_object):

        def add(obj, sla_obj):
            if sla_obj.is_page:
                obj.pages.append(sla_obj)
            else:
                obj.page_objects.append(sla_obj)

            return obj

        if isinstance(sla_object, WireframeObject):
            self = add(self, sla_object)
        else:
            self = add(self, WireframeObject(sla_object))

    def _image_size(self, margins=[0,0]):
        max_x = float()
        max_y = float()

        for po in self.page_objects:
            trx = po.box.coords["top-right"][0].value
            bry = po.box.coords["bottom-right"][1].value

            if trx > max_x:
                max_x = trx

            if bry > max_y:
                max_y = bry

        # We add a little margin

        max_x += margins[0]
        max_y += margins[1]

        return (int(max_x), int(max_y))

    def draw(self, **kwargs):
        """
        Returns Pillow Image instance or bool if [output] option is set.

        :type kwargs: dict
        :param kwargs: Draw options

        **Draw options :**

        +------------------+---------------------------------------+---------------------------+---------+
        | kwargs key       | Use                                   | Type                      | Default |
        +==================+=======================================+===========================+=========+
        | default_outline  | Outline color used if an object has   | boolean or Pillow color   | "black" |
        |                  | no fill and no outline color          |                           |         |
        |                  | defined.                              |                           |         |
        +------------------+---------------------------------------+---------------------------+---------+
        | pages            | Draw all page or only pages in a      | "all" or                  | "all"   |
        |                  | list of page numbers.                 | list of integers [1,...]  |         |
        +------------------+---------------------------------------+---------------------------+---------+
        | layers           | Draw all layers or only layers in     | "all" or                  | "all"   |
        |                  | a list of layer numbers.              | list of integers [1,...]  |         |
        +------------------+---------------------------------------+---------------------------+---------+
        | background_color | Background color of the image         | Pillow color              | "grey"  |
        +------------------+---------------------------------------+---------------------------+---------+
        | output           | File path of the output file          | str                       | False   |
        +------------------+---------------------------------------+---------------------------+---------+
        | stylesheet       | Fill and outline setting according to | boolean or dict           | False   |
        |                  | the type of object to draw.           |                           |         |
        |                  |                                       | (as Wireframe.stylesheet) |         |
        |                  | True for default stylesheet.          |                           |         |
        +------------------+---------------------------------------+---------------------------+---------+
        | landmark         | Draw landmark lines at 0,0.           | boolean                   | True    |
        +------------------+---------------------------------------+---------------------------+---------+
        | bleed            | Draw page bleeds                      | boolean                   | True    |
        +------------------+---------------------------------------+---------------------------+---------+
        """

        draw_pages = "all"
        draw_layers = "all"
        draw_landmark = True
        draw_bleed = True

        out_file = False

        canvas_margins = [0,0]
        background_color = "grey"

        default_outline = "black"

        use_stylesheet = False
        stylesheet = Wireframe.stylesheet

        # --- Options processing ------------------------------------

        for opt_name,opt_value in kwargs.items():

            if opt_name == "layers":
                draw_layers = opt_value

            if opt_name == "pages":

                if isinstance(opt_value, bool):
                    if opt_value:
                        draw_pages = "all"
                    else:
                        draw_pages = "none"
                else:
                    draw_pages = opt_value

            if opt_name == "margins":
                canvas_margins = opt_value

            if opt_name == "background":
                background_color = opt_value

            if opt_name == "default_outline":
                default_outline = opt_value

            if opt_name == "landmark":
                if opt_value:
                    draw_landmark = True
                else:
                    draw_landmark = False

            if opt_name == "bleed":
                if opt_value:
                    draw_bleed = True
                else:
                    draw_bleed = False

            if opt_name == "stylesheet":

                if isinstance(opt_value, bool):
                    if opt_value:
                        use_stylesheet = True
                else:
                    stylesheet = opt_value
                    use_stylesheet = True

            if opt_name == "output":
                out_file = opt_value

        # --- Image creation ----------------------------------------

        image_size = self._image_size(canvas_margins)

        image = Image.new("RGB", image_size, color=background_color)
        canvas = ImageDraw.Draw(image)

        # --- Drawing landmark --------------------------------------

        if draw_landmark:
            canvas.line(((-5,0),(5,0)), fill="red")
            canvas.line(((0,-5),(0,-5)), fill="red")

        # --- Using default_outline or stylesheet -------------------

        if use_stylesheet:
            # --- Using stylesheet ----------------------------------

            for object_set in [self.pages, self.page_objects]:

                for obj in object_set:
                    key = False

                    if obj.type in stylesheet:
                        key = obj.type

                    if key:
                        obj.draw_settings["fill"] = stylesheet[key]["fill"]
                        obj.draw_settings["outline"] = stylesheet[key]["outline"]

        else:
            # --- Using default_outline if no defined outline -------

            if default_outline:

                for object_set in [self.pages, self.page_objects]:

                    for obj in object_set:
                        drawed = False

                        if obj.draw_settings["fill"] or obj.draw_settings["outline"]:
                            drawed = True

                        if not drawed:
                            obj.draw_settings["outline"] = default_outline

            # -------------------------------------------------------

        # --- Drawing page and page objects -------------------------

        if draw_pages != "none":

            if draw_pages == "all":

                for page in self.pages:
                    canvas = page.draw_on_canvas(canvas, draw_bleed)

            else:
                # Only pages in draw_pages list

                for page in self.pages:

                    if page.number in draw_pages:
                        canvas = page.draw_on_canvas(canvas, draw_bleed)

        if draw_layers == "all":

            for pago in self.page_objects:
                canvas = pago.draw_on_canvas(canvas)

        else:
            # Only page objects in draw_layers list layers

            for pago in self.page_objects:

                if pago.layer in draw_layers:
                    canvas = pago.draw_on_canvas(canvas)

        # -----------------------------------------------------------

        if out_file:
            image.save(out_file)

            return True
        else:
            return image

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

"""
Converts an SLA file to PDF with headless Scribus execution.

From Scribus wiki: <https://wiki.scribus.net/canvas/Command_line_scripts>

Usage in Pyscribus:

:Example:

.. code:: python

   from pyscribus.headless import run_pyh_script
   run_pyh_script("topdf", sla_input="/home/user/example.sla")

Usage in command-line:

.. code:: bash

   scribus -g -py to-pdf.py -- file.sla

"""

# Imports ===============================================================#

import sys

from pathlib import Path

# Variables =============================================================#

PATH = __file__
SYNTAX = ["scribus", "-g", "-py", "SCRIPT_PATH", "--", "SLA_INPUT"]

# Program ===============================================================#

# NOTE `scribus` is loaded at Scribus execution.

if __name__ == "__main__":

    if not scribus.haveDoc() :
        print("No file opened.")
        sys.exit(0)

    filepath = Path(scribus.getDocName())

    pdf_filepath = filepath.parent / "{}.pdf".format(filepath.stem)

    pdf = scribus.PDFfile()
    pdf.file = str(pdf_filepath)

    pdf.save()

    sys.exit(0)

# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus classes for stories
"""

# Imports ===============================================================#

# To avoid Sphinx complaints from methods annotations referencing same class
from __future__ import annotations

import copy

from typing import Union, NoReturn, List

import lxml
import lxml.etree as ET

import pyscribus.exceptions as exceptions
import pyscribus.marks as marks
import pyscribus.notes as notes

from pyscribus.common.xml import *

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

class StoryEnding(PyScribusElement):
    """
    Ending marker in Scribus stories (trail)
    """

    def __init__(self, **kwargs):
        super().__init__()
        self.alignment = None
        self.parent = None

    def toxml(self):
        xml = ET.Element("trail")

        if self.alignment is not None:
            xml.attrib["ALIGN"] = self.alignment

        if self.parent is not None:
            xml.attrib["PARENT"] = self.parent

        return xml

    def fromxml(self, xml):
        if xml.tag == "trail":

            align = xml.get("ALIGN")
            if align is not None:
                # TODO use human values...
                self.alignment = align

            parent = xml.get("PARENT")
            if parent is not None:
                self.parent = parent

            return True

        return False


class StoryDefaultStyle(OrphanElement):
    """
    Default style marker in Scribus stories (DefaultStyle)
    """

    def __init__(self):
        OrphanElement.__init__(self, "DefaultStyle")


class StoryLineBreak(OrphanElement):
    """
    Line break marker in Scribus stories (breakline)
    """

    def __init__(self):
        OrphanElement.__init__(self, "breakline")


class NonBreakingHyphen(OrphanElement):
    """
    Non breaking hyphen in Scribus stories (nbhyphen)
    """

    def __init__(self):
        OrphanElement.__init__(self, "nbhyphen")


class NonBreakingSpace(OrphanElement):
    """
    Non breaking space in Scribus stories (nbspace)
    """

    def __init__(self):
        OrphanElement.__init__(self, "nbspace")


class StoryParagraphEnding(PyScribusElement):
    """
    End of paragraph marker in Scribus stories (para)

    :type parent: string
    :param parent: Name of the paragraph style
    :type doc_parent: pyscribus.document.Document
    :param parent: Parent Document instance

    :ivar string parent: Name of the paragraph style
    """

    def __init__(self, parent=False, doc_parent=False):
        super().__init__()

        self.parent = parent

        self.doc_parent = False

    def __repr__(self):
        return "PARAGEND|{}".format(self.parent)

    def fromxml(self, xml, check_style=True):
        parentpar = xml.get("PARENT")
        if parentpar is None:
            self.parent = False
        else:
            self.parent = parentpar

            if check_style and self.parent:

                if self.doc_parent:

                    checked = False

                    for para_style in self.doc_parent.styles["paragraph"]:
                        if para_style.name == self.parent:
                            checked = True
                            break

                    if not checked:
                        raise exceptions.UnknownStyleInStory(self.parent)

        return True

    def toxml(self):
        xml = ET.Element("para")

        if self.parent:
            xml.attrib["PARENT"] = self.parent

        return xml


class StoryFragment(PyScribusElement):
    """
    Text fragment (ITEXT) in Scribus stories.

    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    :ivar string text: Text content
    :ivar dict font: Font details
    :ivar dict features: Font special formatting

    +----------------+---------------------------------+--------------+
    | Kwargs         | Setting                         | Type         |
    +================+=================================+==============+
    | text           | Fragment text                   | string       |
    +----------------+---------------------------------+--------------+
    | font           | Font name                       | string       |
    +----------------+---------------------------------+--------------+
    | fontsize       | Font size                       | float        |
    +----------------+---------------------------------+--------------+
    | fontcolor      | Font color name                 | string       |
    +----------------+---------------------------------+--------------+
    | fontopacity    | Font opacity                    | float        |
    |                |                                 | (percentage) |
    +----------------+---------------------------------+--------------+
    | default        | Equivalent to a fromdefault     | boolean or   |
    |                | call, value being True or the   | string       |
    |                | default name                    |              |
    +----------------+---------------------------------+--------------+
    | features       | Fragment features               | dict         |
    +----------------+---------------------------------+--------------+
    | inherit        | Font feature : inherit          | boolean      |
    +----------------+---------------------------------+--------------+
    | smallcaps      | Font feature : small caps       | boolean      |
    +----------------+---------------------------------+--------------+
    | allcaps        | Font feature : all caps         | boolean      |
    +----------------+---------------------------------+--------------+
    | superscript    | Font feature : superscript      | boolean      |
    +----------------+---------------------------------+--------------+
    | strike         | Font feature : striked          | boolean      |
    +----------------+---------------------------------+--------------+
    | subscript      | Font feature : subscript        | boolean      |
    +----------------+---------------------------------+--------------+
    | underline      | Font feature : underlined       | boolean      |
    +----------------+---------------------------------+--------------+
    | underlinewords | Font feature : underlined words | boolean      |
    +----------------+---------------------------------+--------------+

    :Example:

    .. code:: python

       frag = stories.StoryFragment(
           text="Lorem ipsum", fontsize=7, fontcolor="Grey"
       )

    """

    def __init__(self, **kwargs):
        super().__init__()

        self.text = ""

        # To not export to XML. This is defined by the following
        # StoryParagraphEnding in Story.sequence
        self.paragraph_style = False

        self.font = {
            "name": False,
            "size": False,
            "color": False,
            "opacity": False
        }

        # Features
        self.features = {
            "inherit": False,
            "smallcaps": False,
            "allcaps": False,
            "superscript": False,
            "strike": False,
            "subscript": False,
            "underline": False,
            "underlinewords": False,
        }

        if kwargs:
            self._quick_setup(kwargs)

    def _quick_setup(self, settings):
        """
        Method for defining story fragment settings from class
        instanciation kwargs.

        :type settings: dict
        :param settings: Kwargs dictionnary
        """

        if settings:
            PyScribusElement._quick_setup(self, settings)

            for setting_name, setting_value in settings.items():

                if setting_name == "text":
                    self.text = setting_value

                if setting_value == "features":
                    self.features = setting_value

                if setting_name == "fontsize":
                    self.font["size"] = setting_value

                if setting_name == "fontcolor":
                    self.font["color"] = setting_value

                if setting_name == "fontopacity":
                    self.font["opacity"] = setting_value

                if setting_name == "font":
                    self.font["name"] = setting_value

                if setting_name in self.features.keys():

                    if setting_value:
                        self.features[setting_name] = True
                    else:
                        self.features[setting_name] = False

    def fromdefault(self):
        self.text = ""

        self.paragraph_style = False

        self.font = {
            "name": False,
            "size": False,
            "color": False,
            "opacity": False
        }

        self.features = {
            "inherit": False,
            "smallcaps": False,
            "allcaps": False,
            "superscript": False,
            "strike": False,
            "subscript": False,
            "underline": False,
            "underlinewords": False,
        }

    def __iadd__(self, fragment):
        """
        += operator can be used to join another fragment text.

        The second term **inherits the features of the first**.
        """
        if isinstance(fragment, StoryFragment):
            self.text += fragment.text
            return self
        else:
            raise TypeError()

    def __repr__(self):
        def font_repr(f):
            if f:
                return f
            else:
                return ""

        def features_repr(f):
            return [k for k in f if f[k]]

        return "FRAGMENT|{}|{}|{}".format(
            font_repr(self.font["name"]),
            self.text,
            features_repr(self.features)
        )

    def __str__(self):
        return self.text

    def toxml(self):
        xml = ET.Element("ITEXT")

        have_features = len([f for f in self.features.values() if f])

        if have_features:
            features = []

            for f,v in self.features.items():
                if v:
                    features.append(f)

            features = " ".join(features)

            xml.attrib["FEATURES"] = features

        for case in zip(
                    ["name", "color", "opacity", "size"],
                    ["FONT", "FCOLOR", "FSHADE", "FONTSIZE"],
                    ):
            if self.font[case[0]]:
                xml.attrib[case[1]] = self.font[case[0]]

        xml.attrib["CH"] = self.text

        return xml

    def fromxml(self, xml):
        fragtext = xml.get("CH")
        if fragtext is not None:

            # NOTE Don't do any strip, rstrip to @CH, as it may
            # contains legitimate spaces
            self.text = fragtext

            features = xml.get("FEATURES")
            if features is not None:
                self.set_features(features)

            for case in zip(
                        ["name", "color", "opacity", "size"],
                        ["FONT", "FCOLOR", "FSHADE", "FONTSIZE"],
                        ):

                att = xml.get(case[1])
                if att is not None:
                    self.font[case[0]] = att

            # TODO Reste de l’implémentation, puis :

            return True
        else:
            return False

    def set_features(self, features):
        # TODO

        features = features.split()

        for feature in features:
            if feature in self.features.keys():
                self.features[feature] = True


class StoryVariable(PyScribusElement):
    """
    Variable in a Scribus story (var).

    :type name: string
    :param name: Name of the variable. Must be in StoryVariable.var_names.

    :ivar string name: Name of the variable
    """

    var_names = ["pgno", "pgco"]

    def __init__(self, name=""):
        super().__init__()

        if nam in StoryVariable.var_names or nam == "":
            self.name = name

    def toxml(self):
        """
        :rtype: lxml.etree._Element
        :returns: XML representation of story variable
        """

        element = ET.Element("var")

        element.attrib["name"] = self.name

        return element

    def fromxml(self, xml):
        """
        :rtype: boolean
        :returns: True if XML parsing succeed
        """

        if xml.tag == "var":

            nam = xml.get("name")
            if nam is None:
                return False
            else:
                if nam in StoryVariable.var_names:
                    self.name = nam

            return True

        else:
            return False


class PageNumberVariable(StoryVariable):
    """
    Page number variable in a Scribus story
    """

    def __init__(self):
        super().__init__("pgno")

class PageCountVariable(StoryVariable):
    """
    Count of total pages variable in a Scribus story
    """

    def __init__(self):
        super().__init__("pgco")

# Variables globales 2 ==================================================#

variable_classes = {
    "pgno": PageNumberVariable,
    "pgco": PageCountVariable,
}

# Parsing functions =====================================================#

def sequencefromhtml(
        html,
        font={
            "italic": "Arial Italic",
            "bold": "Arial Bold",
            "bold-italic": "Arial Bold Italic"},
        alternate_emphasis=False,
        sla_document=False):
    """
    :type html: string
    :param html: Paragraph text in PSM (PyScribus Story Markup)
    :type font: dict
    :param font: Fonts used for text emphasis
    :type alternate_emphasis: boolean
    :param alternate_emphasis: Nested emphasis returns to regular font.
    :type sla_document: pyscribus.document.Document
    :param sla_document: Instance of PyScribus Document where notes & marks will
        be stored.
    :rtype: list
    :returns: List of pyscribus.stories.Story sequence elements
    """

    def tag_to_data(fragment, tag):
        """
        Modify fragment font and features according to tag value.
        """

        if tag in ["em", "i"]:
            fragment.font = font["italic"]

        if tag in ["b", "strong"]:
            fragment.font = font["bold"]

        if tag in ["sup"]:
            fragment.features["superscript"] = True

        if tag in ["sub"]:
            fragment.features["subscript"] = True

        if tag in ["u"]:
            fragment.features["underline"] = True

        if tag in ["sc"]:
            fragment.features["smallcaps"] = True

        return fragment

    def attrib_to_data(fragment, element):
        """
        Read @style attribute of element and modify fragment features
        accordingly.
        """

        if element.tag == "span":

            if "style" in element.attrib:
                css = [
                    r.strip()
                    for r
                    in element.attrib["style"].split(";")
                ]

                for rule in css:

                    # Little cleaning -> CSS property name and value
                    rule = rule.replace(": ",":").split(":")
                    name,value = rule[0],":".join(rule[1:])

                    if name == "font-variant":

                        if value == "small-caps":
                            fragment.features["smallcaps"] = True

                    if name == "text-transform":

                        if value == "uppercase":
                            fragment.features["allcaps"] = True

                    if name == "text-decoration":

                        if value == "underline":
                            fragment.features["underline"] = True

        return fragment

    def parse_note(element, sequence, sla_document):
        """
        Parse note XML data wether it comes from a <note> or a
        <span class="note"> element.
        """
        nid = element.get("id")

        if nid is None:
            raise ValueError(
                "<note> without @id."
            )
        else:
            # Mark of the note in the story ----------------------

            story_note = marks.StoryNoteMark(label=nid)
            story_note.fromdefault()

            # Note content in DOCUMENT ---------------------------

            # TODO FIXME Now that the mark is in the story,
            # we must add the note content into the DOCUMENT
            # <Notes>

            if sla_document:
                document_note = notes.Note()
                document_note.fromdefault()
                document_note.parent_mark = nid

                # TODO FIXME Handle note style
                # TODO FIXME Handle note text

                sla_document.notes.append(document_note)

            # Mark of the note in the DOCUMENT marks -------------

            if sla_document:
                document_mark = marks.DocumentMark()
                document_mark.fromdefault("note")
                document_mark.label = nid
                sla_document.marks.append(document_mark)

            # ----------------------------------------------------

            sequence.append(story_note)

        return sequence, sla_document

    def parse_element(element, sequence, font, previous, alternate_emphasis, sla_document):
        global variable_classes

        if element.tag in ["em", "i", "span", "b", "strong", "sup", "sub", "u", "sc"]:
            is_fragment = True

        if element.tag == "note":
            is_fragment = False

            sequence, sla_document = parse_note(element, sequence, sla_document)

        if element.tag in ["pgno", "pgco"]:
            is_fragment = False

            variable = Story.variable_classes[element.tag]()
            sequence.append(variable)

        if element.tag == "br":
            is_fragment = False
            sequence.append(StoryLineBreak())

        if element.tag == "span":

            if "class" in element.attrib:
                element_classes = [
                    cn.strip()
                    for cn in element.attrib["class"].strip().split()
                ]

                # Get the count of incompatible classes in span @class attribute
                # So <span class="pgno note"></span> will not work.
                main_class = len(
                    [
                        c
                        for c in ["pgno", "pgco", "note"]
                        if c in element_classes
                    ]
                )

                if main_class <= 1:

                    for element_class in element_classes:
                        # NOTE This iteration to check multiple classes in @class
                        # attribute, if a precision of a class is needed.

                        if element_class in ["pgno", "pgco"]:
                            is_fragment = False

                            variable = Story.variable_classes[element_class]()
                            sequence.append(variable)

                            break

                        if element_class == "note":
                            is_fragment = False

                            sequence, sla_document = parse_note(
                                element, sequence, sla_document
                            )

                            break
                else:
                    raise ValueError(
                        "Incompatible classes in span element @class : '{}'".format(
                            ",".join(element_classes)
                        )
                    )

        if is_fragment:
            fragment = StoryFragment(text=element.text)

            read_tag = True

            if alternate_emphasis:
                # If <em>lorem <em>ipsum</em> et dolor</em> must
                # produce an "ipsum" in regular style surrounded
                # by "lorem", "et dolor" both in italic.
                # Same thing for bold.

                if previous.tag in ["b", "strong", "i", "em"]:
                    if element.tag in ["b", "strong", "i", "em"]:
                        read_tag = False

            if read_tag:
                fragment = tag_to_data(fragment, element.tag)

            fragment = attrib_to_data(fragment, element)

            sequence.append(fragment)

        if len(element):

            for sub_element in element:

                sequence = parse_element(
                    sub_element,
                    sequence,
                    font,
                    element,
                    alternate_emphasis,
                    sla_document
                )

        if element.tail is not None:
            tail_fragment = StoryFragment(text=element.tail)
            tail_fragment = tag_to_data(tail_fragment, previous.tag)
            tail_fragment = attrib_to_data(tail_fragment, previous)

            sequence.append(tail_fragment)

        return sequence

    global variable_classes

    # Prepare parsing -----------------------------------------------

    # Encapsulate html in a <p> element if it doesn't exists, to
    # make lxml parsing work.

    if not html.startswith("<p>"):
        html = "<p>{}".format(html)

    if not html.endswith("</p>"):
        html = "{}</p>".format(html)

    parsed = ET.fromstring(html)

    # Parsing--------------------------------------------------------

    sequence = []

    if parsed.text is not None:
        sequence.append(StoryFragment(text=parsed.text))

    if len(parsed):

        for element in parsed:

            sequence = parse_element(
                element,
                sequence,
                font,
                parsed,
                alternate_emphasis,
                sla_document
            )

    if parsed.tail is not None:
        sequence.append(StoryFragment(text=parsed.tail))

    # ---------------------------------------------------------------

    return sequence

# Story class ===========================================================#

class Story(PyScribusElement):
    """
    Story in SLA (StoryText). A text buffer.

    :type sla_parent: pyscribus.sla.SLA
    :param sla_parent: Parent SLA instance.
    :type doc_parent: pyscribus.document.Document
    :param doc_parent: Parent Document instance.
    :type pgo_parent: pyscribus.pageobjects.PageObject
    :param pgo_parent: Parent page object instance.

    :ivar pyscribus.sla.SLA sla_parent: Parent SLA instance.
    :ivar pyscribus.document.Document doc_parent: Parent Document instance.
    :ivar list sequence: List of instances of StoryFragment,
        StoryParagraphEnding, StoryLineBreak, StoryDefaultStyle,
        NonBreakingHyphen, StoryVariable
    """

    def __init__(self, sla_parent=False, doc_parent=False, pgo_parent=False):
        super().__init__()

        self.sequence = []

        self.sla_parent = sla_parent
        self.doc_parent = doc_parent
        self.pgo_parent = pgo_parent

    def _without_ending(self):
        temp = []

        if self.sequence:
            if isinstance(self.sequence[-1], StoryEnding):
                temp = self.sequence[:-1]

        return temp

    def append_paragraph(self, **kwargs):
        """
        Append a paragraph at the end of the story.

        :type kwargs: dict
        :param kwargs: kwargs

        +-------------------+--------------+---------+--------------------------+
        | Kwargs key        | Kwargs value | Default |Usage                     |
        +===================+==============+=========+==========================+
        |text               | string       | False   | Paragraph text in PSM    |
        |                   |              |         | (PyScribus Story Markup) |
        +-------------------+--------------+---------+--------------------------+
        |ending             | boolean      | True    | Add a paragraph ending   |
        +-------------------+--------------+---------+--------------------------+
        |inherit_style      | boolean      | False   | Inherit paragraph style  |
        |                   |              |         | from last paragraph      |
        +-------------------+--------------+---------+--------------------------+
        |style              | string       | False   | Style of the paragraph.  |
        |                   |              |         | Overrides inherit_style. |
        +-------------------+--------------+---------+--------------------------+
        |alternate_emphasis | boolean      | False   | Nice emphasis, as nested |
        |                   |              |         | italic in italic becames |
        |                   |              |         | regular. Same for bold   |
        |                   |              |         | nested in bold.          |
        +-------------------+--------------+---------+--------------------------+

        As there is only one paragraph added by append_paragraph, the <p> element,
        enclosing the paragraph, can be omitted in text argument.

        :Example:

        .. code:: python

           story.append_paragraph(
               text="Le chant du cygne est très beau.", style="Normal"
           )
           story.append_paragraph(
               text="- Où as-tu <em>vu</em> ça ?", inherit_style=True
           )
           story.append_paragraph(
               text="<p>Et c'est ainsi qu'ils se quittèrent</p>",
               ending=False
           )

        """

        self._append_paragraph(kwargs)

        return True

    def append_paragraphs(self, paragraphs=[]):
        """
        Append multiple paragraphs at the end of the story.

        :type paragraphs: list
        :param paragraphs: List of Story.append_paragraph() kwargs dictionnaries.

        :Example:

        .. code:: python

           story.append_paragraphs(
              [
                  {"text": "A title", "style": "Title1"},
                  {"text": "Foreword", "style": "Foreword"},
                  {
                      "text": "First paragraph of content",
                      "style": "Normal"
                  },
                  {
                      "text": "Second paragraph of content",
                      "inherit_style": True
                  },
                  {"text": "03/01/2020", "ending": False}
              ]
           )

        .. seealso:: pyscribus.stories.Story.append_paragraph
        """

        if paragraphs:

            for paragraph in paragraphs:
                self._append_paragraph(paragraph)

            return True
        else:
            return False

    def _append_paragraph(self, kwargs):
        """
        Private method to append paragraph.

        Used by append_paragraph() and append_paragraphs()

        :type kwargs: dict
        :param paragraphs: kwargs of Story.append_paragraph()
        """

        # kwargs processing -------------------------------

        style,source,inherit_style = False,False,False
        alternate_emphasis = False

        ending = True

        font={
            "bold": "Arial Bold",
            "italic": "Arial Italic",
            "bold-italic": "Arial Bold Italic"
        }

        for param,value in kwargs.items():

            if param == "text":
                source = value

            if param == "ending":
                ending = bool(value)

            if param == "inherit_style":
                inherit_style = bool(value)

            if param == "style":
                style = value

            if param == "alternate_emphasis":
                alternate_emphasis = bool(value)

            if param == "font":

                if isinstance(value, dict):

                    for font_style in ["bold", "italic", "bold-italic"]:

                        if font_style not in value:

                            raise ValueError(
                                "Missing {} setting to font param".format(
                                    font_style
                                )
                            )

                    font = value

        # -------------------------------------------------

        sequence = sequencefromhtml(
            source,
            font=font,
            alternate_emphasis=alternate_emphasis,
            sla_document=self.doc_parent
        )

        if ending:
            para = StoryParagraphEnding()

            if style or inherit_style:
                ps = False

                # NOTE style overrides the inherited style

                if inherit_style:
                    # Get the last paragraph ending style

                    ps = [
                        e.parent
                        for e in self.sequence
                        if isinstance(e, StoryParagraphEnding)
                    ][-1]

                if style:
                    ps = style

                para.parent = ps

            sequence.append(para)

        temp = self._without_ending()
        self.sequence = temp + sequence
        self.end_contents()

    def toxml(self):
        """
        Return the story as XML element.

        :rtype: lxml._Element
        """

        xml = ET.Element("StoryText")

        for element in self.sequence:
            ex = element.toxml()
            xml.append(ex)

        return xml

    def fromxml(self, xml, check_style=True):
        """
        Parses XML of a SLA Story.

        :type xml: lxml._Element
        :param xml: SLA Story as lxml._Element
        :type check_style: bool
        :param check_style: Check if story paragraphs use known paragraph
            styles of Story.doc_parent Document. True by default.

        .. note:: You might set check_style to False if you want to parse
            stories independantly of any given SLA & Document instances.

        :rtype: bool
        :returns: bool

        """

        global variable_classes

        if xml.tag == "StoryText":

            for element in xml:

                if element.tag == "DefaultStyle":
                    self.sequence.append(StoryDefaultStyle())

                if element.tag == "ITEXT":

                    frag = StoryFragment()
                    success = frag.fromxml(element)

                    if success:
                        self.sequence.append(frag)

                # End of a paragraph

                if element.tag == "para":
                    para = StoryParagraphEnding(self.doc_parent)
                    success = para.fromxml(element, check_style)

                    if success:

                        # NOTE @PARENT carries the style name of the preceding
                        # StoryFragment / <ITEXT> in the sequence

                        if para.parent:

                            # NOTE Empty lines in the story are just <para>
                            # following each others. So we only set the
                            # paragraph_style of the preceding fragment…
                            # if there *is* a fragment to begin with.

                            if isinstance(self.sequence[-1], StoryFragment):
                                self.sequence[-1].paragraph_style = para.parent

                        self.sequence.append(para)

                # Line break in a paragraph

                if element.tag == "breakline":
                    self.sequence.append(StoryLineBreak())

                # Non breaking hyphen

                if element.tag == "nbhyphen":
                    self.sequence.append(NonBreakingHyphen())

                # Non breaking space

                if element.tag == "nbspace":
                    self.sequence.append(NonBreakingSpace())

                # Mark

                if element.tag == "MARK":
                    # TODO Selon @type, un StoryNoteMark ou autre
                    pass

                # Variable

                if element.tag == "var":

                    var_name = element.get("name")

                    if var_name is not None:
                        var_instance = Story.variable_classes[var_name]()
                        self.sequence.append(var_instance)

                # Story end

                if element.tag == "trail":
                    story_ending = StoryEnding()
                    success = story_ending.fromxml(element)

                    if success:
                        self.sequence.append(story_ending)
                    break

            return True

        return False

    def fromdefault(self):
        """
        Set a default story
        """

        self.sequence = [StoryDefaultStyle(), StoryEnding()]

    def append(self, element):
        """
        Append a element to the Story sequence.

        .. warning:: Don't use append() to append a StoryEnding, use
            end_contents() instead.

        :type element: StoryFragment, StoryParagraphEnding, StoryLineBreak, StoryDefaultStyle
        :param element: Element to add to the story sequence
        :rtype: bool
        :returns: bool

        .. seealso:: pyscribus.stories.end_contents
        """

        temp = self._without_ending()
        self.sequence = temp + [element]
        self.end_contents()

        return True

    def end_contents(self, no_trailing_paragraph=False):
        """
        Ends the story's content by making sure that the story sequence
        is valid.

        .. note:: It is better to use this method than appending a
            :class:`StoryEnding` instance directly into the story sequence.

        :type no_trailing_paragraph: bool
        :param no_trailing_paragraph: Remove last paragraph ending
        """
        story_ending = None

        if isinstance(self.sequence[-1], StoryEnding):
            story_ending = self.sequence.pop(-1)

        if no_trailing_paragraph:
            if isinstance(self.sequence[-1], StoryParagraphEnding):
                self.sequence.pop(-1)

        if story_ending is None:
            self.sequence.append(StoryEnding())
        else:
            self.sequence.append(story_ending)

    def init_contents(self):
        self.sequence = [StoryDefaultStyle()]

    def bypars(self):
        """
        Return a list of list of Story.sequence elements organized as human
        paragraphs.

        :returns: list of lists
        :rtype: ilst
        """

        pars = [] # List of paragraphs
        tmp = [] # Temporary list of paragraph elements

        for element in self.sequence:

            # End of the story -> last paragraph is added to the list

            if isinstance(element, StoryEnding):
                if tmp:
                    pars.append(tmp)
                    break

            # End of a paragraph ------------------------------------

            if isinstance(element, StoryParagraphEnding):

                if tmp:
                    tmp.append(element)
                    pars.append(tmp)
                    tmp = []

            # Paragraph content -------------------------------------

            for class_test in [StoryFragment, StoryDefaultStyle,
                    NonBreakingSpace, NonBreakingHyphen, StoryVariable]:

                if isinstance(element, class_test):
                    tmp.append(element)
                    break

        return pars

    def rawtext(self):
        """
        Returns a text string equivalent to *what Scribus story editor
        saves* as txt file.

        :rtype: string
        """
        text = ""

        for par in self.bypars():

            for element in par:

                # NOTE Text formatting is not saved
                if isinstance(element, StoryFragment):
                    text += element.text

                if isinstance(element, NonBreakingSpace):
                    text += " "

                # NOTE Line breaks are exported by Scribus... as spaces
                if isinstance(element, StoryLineBreak):
                    text += " "

                if isinstance(element, StoryParagraphEnding):
                    text += "\n"

        return text

    def templatable(self):
        """
        Return elements of Story sequence that are available for templating.

        :rtype: list
        :returns: List of pyscribus.stories.StoryFragment
        """

        contents = []

        pattern = self.sla_parent.templating["intext-pattern"]

        for element in self.sequence:

            if isinstance(element, StoryFragment):

                if pattern.findall(element.text):
                    contents.append(element)

        return contents

    def feed_templatable(self, datas: dict) -> Union[bool, List[StoryFragment]]:
        """
        Feed template-able datas into this story.

        :returns: List of modified elements in this story, or False
        :rtype: Union[bool, list]
        """

        elements = self.templatable()

        if elements:
            modified = []

            for element in elements:

                for template_key, template_value in datas.items():

                    if template_key not in element.text:
                        continue

                    new_value = None

                    if isinstance(datas[template_key], list):
                        # We have %key% -> ["Value1", "Value2", …]
                        # So the first occurrence of %key% is "Value1", the
                        # second "Value2", etc.

                        if len(datas[template_key]):
                            new_value = datas[template_key].pop(0)
                    else:
                        # We have %key% -> "Value"
                        new_value = template_value

                    if new_value is None:
                        continue

                    element.text = element.text.replace(
                        template_key, new_value
                    )

                    modified.append(element)

            return modified

        return False

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus classes for measures and geometrical manipulations.
"""

import copy
import math

from typing import Union, Literal

from pyscribus.common.math import PICA_TO_MM

import pyscribus.logs as logs
import pyscribus.exceptions as exceptions

import pyscribus.papers.ansi as ansipaper
import pyscribus.papers.iso216 as iso216paper

from pyscribus.common.xml import *

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

XY = Literal["x", "y"]
DimValue = Union[float, int]
BoxCorner = Literal["top-left", "top-right", "bottom-left", "bottom-right"]

# Classes ===============================================================#

class Dim:
    """
    Dimension object. Allows to convert and export in the correct unit.

    :type value: Union[float, int]
    :param value: Value of the dimension

    :type unit: str
    :param unit: Unit of the dimension.
    :type is_int: boolean
    :param is_int: Use integer instead of float type for value

    :type kwargs: dict
    :param kwargs: Dimension settings (see kwargs table)

    :ivar float,int value: Value of the dimension in the original unit
    :ivar string unit: Unit of the dimension

    +---------------+------------------------------------+
    | Kwargs        | Setting                            |
    +===============+====================================+
    | unit          | Unit of the dimension. See unit    |
    |               | table.                             |
    +---------------+------------------------------------+
    | integer       | The value of the dimension is an   |
    |               | integer instead of a float.        |
    +---------------+------------------------------------+
    | negative      | Allow the dimension to have        |
    |               | negative value.                    |
    +---------------+------------------------------------+
    | original_unit |                                    |
    +---------------+------------------------------------+

    +--------------------------+---------------+
    | Unit / Notation          | unit argument |
    +==========================+===============+
    | Milimeter                | mm            |
    +--------------------------+---------------+
    | Pica point               | pica, pct     |
    +--------------------------+---------------+
    | Typographic point        | pt            |
    +--------------------------+---------------+
    | Percentage (0 to 100)    | perc, pc      |
    +--------------------------+---------------+
    | Percentage (0.0 to 1)    | pcdecim, pcd  |
    +--------------------------+---------------+
    | Percentage (decimal, can | pcscale       |
    | be more than 1.0)        |               |
    +--------------------------+---------------+
    | Calligraphic pen degree  | cdeg          |
    +--------------------------+---------------+
    | Regular degree           | deg           |
    +--------------------------+---------------+
    | Dot per inch (DPI/PPP)   | dpi, ppp, ppi |
    +--------------------------+---------------+
    | Line per inch (LPI)      | lpi           |
    +--------------------------+---------------+
    | Second                   | s, sec        |
    +--------------------------+---------------+
    """

    # 1 pica point = 25,4/72 milimeters
    PICA_TO_MM = (25.4 / 72)

    UNIT_ARGS = {
        "mm": ["mm"],
        "pica": ["pica", "pct"],
        "pt": ["pt"],
        "perc": ["perc", "pc"],
        "pcdecim": ["pcdecim", "pcd"],
        "cdeg": ["cdeg"],
        "deg": ["deg"],
        "dpi": ["dpi", "ppp", "ppi"],
        "lpi": ["lpi"],
        "sec": ["s", "sec"]
    }

    def __init__(self, value: DimValue, **kwargs: dict):
        self.value = value
        self.integer = False
        self.allow_negative = False

        if "integer" in kwargs:
            self.integer = bool(kwargs["integer"])

        if "negative" in kwargs:
            self.allow_negative = bool(kwargs["negative"])

        if self.integer:
            self.value = int(self.value)

        if "unit" in kwargs:
            self.set_unit(kwargs["unit"])
        else:
            self.set_unit("pica")

        if "original_unit" in kwargs:
            if kwargs["original_unit"]:
                self.from_original_unit(kwargs["original_unit"])

        print(self.__repr__())

        self.check_value()

    #--- Checking, setting the unit --------------------------------------

    def from_original_unit(self, original_unit):
        pass

    def check_value(self):
        """
        Check value validity according to unit.

        :rtype: boolean
        :return: True if the value is valid
        :raises pyscribus.exceptions.InvalidDim: Raised if value is invalid.
        """

        if self.unit in ["pica", "pct"]:

            if self.value < 0:
                if not self.allow_negative:
                    raise exceptions.InvalidDim(
                        "Pica points must not be inferior to 0 in that context. From {}".format(
                            self.__repr__()
                        )
                    )

        if self.unit == "dpi":

            if self.value < 0:
                raise exceptions.InvalidDim(
                    "DPI/PPP must be a positive number"
                )

        if self.unit == "lpi":

            if self.value < 0:
                raise exceptions.InvalidDim(
                    "LPI must be a positive number"
                )

        if self.unit == "sec":
            if int(float(self.value)) != float(self.value):
                raise exceptions.InvalidDim(
                    "Second Dim must be an integer, is {}".format(self.value)
                )

        if self.unit == "cdeg":

            if self.value >= 0:
                if self.value <= 180:
                    return True
                else:
                    raise exceptions.InvalidDim(
                        "Calligraph pen angle must range from 0 to 180"
                    )
            else:
                raise exceptions.InvalidDim(
                    "Calligraph pen angle must range from 0 to 180"
                )

        if self.unit == "deg":

            if self.value >= 0:

                if self.value <= 360:
                    return True
                else:
                    raise exceptions.InvalidDim(
                        "Angle must range from 0 to 360"
                    )

            else:
                raise exceptions.InvalidDim(
                    "Angle must range from 0 to 360"
                )

        return True

    def set_unit(self, unit="pica"):
        """
        Set the unit used.

        :type value: float,int
        :param value: Value of the dimension
        :type unit: str
        :param unit: Unit of the dimension.
        :rtype: boolean
        :return: True if the unit is valid

        +-------------------------+---------------+
        | Unit / Notation         | unit argument |
        +=========================+===============+
        | Milimeter               | mm            |
        +-------------------------+---------------+
        | Pica point              | pica, pt      |
        +-------------------------+---------------+
        | Percentage (0 to 100)   | perc, pc      |
        +-------------------------+---------------+
        | Percentage (0.0 to 1,   | pcdecim, pcd  |
        | also negative)          |               |
        +-------------------------+---------------+
        | Calligraphic pen degree | cdeg          |
        +-------------------------+---------------+
        | Regular degree          | deg           |
        +-------------------------+---------------+
        | Dot per inch (DPI/PPP)  | dpi, ppp, ppi |
        +-------------------------+---------------+
        | Line per inch (LPI)     | lpi           |
        +-------------------------+---------------+
        | Second                  | s, sec        |
        +-------------------------+---------------+
        """

        tmp_unit = unit.lower()
        valid_unit = False

        for code,args in Dim.UNIT_ARGS.items():

            if tmp_unit in args:
                self.unit = code
                valid_unit = True
                break

        if self.unit == "sec":
            self.integer = True
            self.value = int(self.value)

        return valid_unit

    #--- XML export ------------------------------------------------------

    def toxmlstr(self, no_useless_decimals=False):
        """
        Returns a XML string of the dimension according to its unit.

        :type no_useless_decimals: boolean
        :param no_useless_decimals: Returns integer instead of float if
            decimals are 0. So 1.0 -> 1 ; 1.1 -> 1.1.
        :rtype: str
        :return: str
        """

        def decimals(n):
            if float(n) == int(n):
                return int(n)
            else:
                return n

            return number

        if self.unit not in [
            "perc",
            "pcdecim",
            "pcscale",
            "cdeg",
            "deg",
            "lpi",
            "dpi",
            "pt",
        ]:
            if no_useless_decimals:
                pica = self.topica()
                return str(decimals(pica))
            else:
                return str(self.topica())
        else:
            if self.unit == "pcdecim":
                if self.value == 1.0:
                    return "1"
                else:
                    if no_useless_decimals:
                        return str(decimals(self.value))
                    else:
                        return str(self.value)
            else:
                if no_useless_decimals:
                    return str(decimals(self.value))
                else:
                    return str(self.value)

    def toxml(self):
        """
        Alias of Dimension.toxmlstr()

        Returns a XML string of the dimension according to its unit.

        :rtype: str
        :return: str
        """

        return self.toxmlstr()

    #--- Conversion into other units -------------------------------------

    def _ceil(self, value, ceil=False):
        """
        Ceil value if ceil is True.

        :type value: float,int
        :param value: Value to ceil (or not)
        :type ceil: boolean
        :param ceil: If True, apply ceil to value
        :rtype: float,int
        :return: Ceiled (or not) number
        """

        if ceil:
            return math.ceil(value)
        else:
            return value

    def _convertorval(self, obj, value, unit):
        """
        Returns a Dim object with unit <unit> if <obj>, or <value>.

        :rtype: float,int,Dim
        :return: Dim object or Dim value
        """

        if obj:
            return Dim(value, unit=unit)

        return value

    def is_convertable_length(self):
        """
        :rtype: boolean
        :return: If the Dim instance is a convertable length
        """

        return not (
            self.unit
            in ["perc", "pcdecim", "pcscale", "cdeg", "deg", "dpi", "lpi"]
        )

    def topica(
        self, ceil: bool = False, obj: bool = False
    ) -> Union[float, int, Dim]:
        """
        Returns the value of Dim in pica point unit.
        Raise ValueError if Dim is not convertable

        :type ceil: boolean
        :param ceil: If True, apply ceil to returned value
        :type obj: boolean
        :param obj: If True, returns a Dim object instead of value.
        """

        if not self.is_convertable_length():
            raise exceptions.IncompatibleDim(
                "Can't convert that ({}) into pica points".format(self.unit)
            )

        if self.unit == "pica":
            return self._convertorval(
                obj, self._ceil(self.value, ceil), "pica"
            )

        if self.unit == "mm":
            return self._convertorval(
                obj, self._ceil(self.value / Dim.PICA_TO_MM, ceil), "pica"
            )

    def todpi(
        self, ceil: bool = False, obj: bool = False
    ) -> Union[float, int, Dim]:
        """
        Returns the value of Dim in DPI unit.
        Raise ValueError if Dim is not convertable

        :type ceil: boolean
        :param ceil: If True, apply ceil to returned value
        :type obj: boolean
        :param obj: If True, returns a Dim object instead of value.
        """

        if self.unit == "lpi":
            return self._convertorval(
                obj, self._ceil(self.value * 16), "dpi"
            )
        else:
            raise exceptions.IncompatibleDim(
                "Can't convert that into DPI"
            )

    def topc(self, ceil, obj=False):
        """
        Returns the value of Dim as integer percentage.
        Raise ValueError if Dim is not convertable

        :type ceil: boolean
        :param ceil: If True, apply ceil to returned value
        :type obj: boolean
        :param obj: If True, returns a Dim object instead of value.
        """

        if self.unit == "pcdecim":
            return self._convertorval(
                obj, self._ceil(self.value * 100), "pc"
            )
        else:
            raise exceptions.IncompatibleDim(
                "Can't convert that into percentage"
            )

    def tolpi(
        self, ceil: bool = False, obj: bool = False
    ) -> Union[float, int, Dim]:
        """
        Returns the value of Dim in LPI unit.
        Raise ValueError if Dim is not convertable

        :type ceil: boolean
        :param ceil: If True, apply ceil to returned value
        :type obj: boolean
        :param obj: If True, returns a Dim object instead of value.
        """

        if self.unit == "dpi":
            return self._convertorval(
                obj, self._ceil(self.value / 16), "lpi"
            )
        else:
            raise exceptions.IncompatibleDim(
                "Can't convert that into LPI"
            )

    def tomm(
        self, ceil: bool = False, obj: bool = False
    ) -> Union[float, int, Dim]:
        """
        Returns the value of Dim in milimeter unit.

        :type ceil: boolean
        :param ceil: If True, apply ceil to returned value
        :type obj: boolean
        :param obj: If True, returns a Dim object instead of value.
        """

        if not self.is_convertable_length():
            raise exceptions.IncompatibleDim(
                "Can't convert that into milimeters"
            )

        if self.unit == "pica":
            return self._convertorval(
                obj, self._ceil(self.value * Dim.PICA_TO_MM, ceil), "mm"
            )

        if self.unit == "mm":
            return self._convertorval(
                obj, self._ceil(self.value, ceil), "mm"
            )

    #--- Defaults --------------------------------------------------------

    def fromdefault(self, default):
        """
        Set Dim attributes according to a named default.

        :type default: str
        :param default: Name of the set of defaults.
        :rtype: boolean
        :return: boolean
        """

        if default.startswith("a4-"):
            self.set_unit("pica")

            if default == "a4-width":
                self.value = iso216paper.A4.WIDTH

            if default == "a4-height":
                self.value = iso216paper.A4.HEIGHT

        if default.startswith("letter-"):
            self.set_unit("pica")

            if default == "letter-width":
                self.value = ansipaper.Letter.WIDTH

            if default == "letter-height":
                self.value = ansipaper.Letter.HEIGHT

        return True

    #--- Python __ methods -----------------------------------------------

    def __bool__(self):
        return bool(self.value)

    def __str__(self):
        r,u = str(self.value),""

        if self.unit in Dim.UNIT_ARGS.keys():
            short = {"mm": "mm", "perc": "%", "pica": "pct", "lpi": "lpi"}

            if self.unit in short:
                u = short[self.unit]

            if self.unit in ["cdeg", "deg"]:
                u = "°"

        else:
            raise exceptions.UnknownDimUnit(self.unit)

        return "{} {}".format(r, u)

    def __repr__(self):
        return "Dim(value={}, unit={}, integer={}, negative={})".format(
            self.value, self.unit, self.integer, self.allow_negative
        )

    def __float__(self):
        return float(self.value)

    def __int__(self):
        return int(self.value)

    def __iadd__(self, dim):

        if isinstance(dim, float):
            self.value += dim

        if isinstance(dim, Dim):
            if dim.unit == self.unit:
                self.value += dim.value
            else:
                raise exceptions.IncompatibleDim()

        return self

    def __isub__(self, dim):

        if isinstance(dim, float):
            self.value -= dim

        if isinstance(dim, Dim):
            if dim.unit == self.unit:
                self.value -= dim.value
            else:
                raise exceptions.IncompatibleDim()

        return self

    def __imul__(self, dim):

        if isinstance(dim, float):
            self.value *= dim

        if isinstance(dim, Dim):
            if dim.unit == self.unit:
                self.value *= dim.value
            else:
                raise exceptions.IncompatibleDim()

    def __sub__(self, dim):
        return self.__isub__(dim)

    def __add__(self, dim):
        return self.__iadd__(dim)

    def __mul__(self, dim):
        return self.__imul__(dim)


class DimBox:
    """
    Box/rectangle object to manipulate Scribus frames coordinates.

    :type kwargs: dict
    :param kwargs: kwargs (see kwargs table)

    :ivar dict dims: Width and height of the box as Dim objects
    :ivar dict coords: Coordinates of the box, dict of list of
        Dim objects ([x, y]) for each point
    :ivar Dim rotation: Rotation angle of the box as Dim object
        (unit : degree)
    :ivar dict rotated_coords: Coordinates of the box when rotated by
        rotation degree

    +-------------------------+------------+
    | Box point coordinate    | kwargs key |
    +=========================+============+
    | Top left x position     | top_lx     |
    +-------------------------+------------+
    | Top left y position     | top_ly     |
    +-------------------------+------------+
    | Top right x position    | top_rx     |
    +-------------------------+------------+
    | Top right y position    | top_ry     |
    +-------------------------+------------+
    | Bottom right x position | bottom_rx  |
    +-------------------------+------------+
    | Bottom right y position | bottom_ry  |
    +-------------------------+------------+
    | Box width               | width      |
    +-------------------------+------------+
    | Box height              | height     |
    +-------------------------+------------+
    """

    def __init__(self,**kwargs):
        # X,Y coordinates for each corner
        # Use setx, sety, getx, gety methods as shorthands
        self.coords = {
            "top-left": [Dim(0), Dim(0)],
            "top-right": [Dim(0), Dim(0)],
            "bottom-left": [Dim(0), Dim(0)],
            "bottom-right": [Dim(0), Dim(0)],
        }

        # Height and width
        self.dims = {"width": Dim(0), "height": Dim(0)}

        # Coords & angle for rotated boxes --------------------------
        # NOTE rotated_coords is modified through set_box and rotate

        self.rotation = Dim(0, unit="deg")
        self.rotated_coords = {
            "top-left": [Dim(0), Dim(0)],
            "top-right": [Dim(0), Dim(0)],
            "bottom-left": [Dim(0), Dim(0)],
            "bottom-right": [Dim(0), Dim(0)],
        }

        # -----------------------------------------------------------

        self.set_box(kwargs=kwargs)

    #--- Shorthands for corners coordinates ------------------------------

    def _setxy(
        self, corner: BoxCorner, value, xy: XY, rotated: bool = False
    ) -> bool:
        if corner.lower() in [
                "top-left", "top-right",
                "bottom-left", "bottom-right"]:

            if xy == "x":
                if rotated:
                    self.rotated_coords[corner][0].value = value
                else:
                    self.coords[corner][0].value = value
            else:
                if rotated:
                    self.rotated_coords[corner][1].value = value
                else:
                    self.coords[corner][1].value = value

            return True
        else:
            raise KeyError()

    def _getxy(
        self, corner: BoxCorner, xy: XY, rotated: bool = False
    ) -> Union[float, int]:
        if corner.lower() in [
                "top-left", "top-right",
                "bottom-left", "bottom-right"]:
            if xy == "x":
                if rotated:
                    return self.rotated_coords[corner][0].value
                else:
                    return self.coords[corner][0].value
            else:
                if rotated:
                    return self.rotated_coords[corner][1].value
                else:
                    return self.coords[corner][1].value
        else:
            raise KeyError()

    def setx(self, corner, value, rotated=False):
        return self._setxy(corner, value, "x", rotated)

    def sety(self, corner, value, rotated=False):
        return self._setxy(corner, value, "y", rotated)

    def getx(
        self, corner: BoxCorner, rotated: bool = False
    ) -> Union[float, int]:
        return self._getxy(corner, "x", rotated)

    def gety(
        self, corner: BoxCorner, rotated: bool = False
    ) -> Union[float, int]:
        return self._getxy(corner, "y", rotated)

    #--- Box modification ------------------------------------------------

    def move(
        self,
        posx: float = 0,
        posy: float = 0,
        refpoint: BoxCorner = "top-left",
    ) -> bool:
        """
        Move the box at posx, posy position.

        :type posx: float
        :param posx: New X position
        :type posy: float
        :param posy: New Y position
        :type refpoint: string
        :param refpoint: Coordinate point of reference (DimBox.coords key)
        :rtype: boolean
        """

        origin_x,origin_y = None,None

        if refpoint in self.coords.keys():
            origin_x = self.coords[refpoint][0].value
            origin_y = self.coords[refpoint][1].value

        if origin_x is not None and origin_y is not None:

            if origin_x != posx or origin_y != posy:
                # Noving the box according to the reference point

                if refpoint == "top-left":
                    self.set_box(
                        top_lx=posx, top_ly=posy,
                        width=self.dims["width"].value,
                        height=self.dims["height"].value,
                    )

                    return True

            else:
                # Moving the box at the exact same position is not a mistake
                # even if it's useless
                return True

        return False

    def translate(self, amountx=0, amounty=0, refpoint="top-left"):
        """
        Move the box by an amount of amountx, amounty

        :type amountx: float
        :param amountx: Amount of X translation
        :type amounty: float
        :param amounty: Amount of Y translation
        :type refpoint: string
        :param refpoint: Coordinate point of reference (DimBox.coords key)
        :rtype: boolean
        """

        origin_x,origin_y = None,None

        if refpoint in self.coords.keys():
            origin_x = self.coords[refpoint][0].value
            origin_y = self.coords[refpoint][1].value

        if origin_x is not None and origin_y is not None:

            # Translating the box by amountx, amounty

            if refpoint == "topleft":
                npx = self.coords["top-left"][0].value + amountx
                npy = self.coords["top-left"][1].value + amounty

                self.set_box(
                    top_lx=npx, top_ly=npy,
                    width=self.dims["width"].value,
                    height=self.dims["height"].value,
                )

                return True

        return False

    def set_box(self, **kwargs):
        """
        Set all coordinates of the box from a set a coordinates
        and/or width & height.

        +-------------------------+------------+
        | Box point coordinate    | kwargs key |
        +=========================+============+
        | Top left x position     | top_lx     |
        +-------------------------+------------+
        | Top left y position     | top_ly     |
        +-------------------------+------------+
        | Top right x position    | top_rx     |
        +-------------------------+------------+
        | Top right y position    | top_ry     |
        +-------------------------+------------+
        | Bottom right x position | bottom_rx  |
        +-------------------------+------------+
        | Bottom right y position | bottom_ry  |
        +-------------------------+------------+
        | Box width               | width      |
        +-------------------------+------------+
        | Box height              | height     |
        +-------------------------+------------+

        :rtype: boolean
        """

        def all_case(obj, kwargs):
            """
            Define DimBox points from all corners.

            X-------X
            |       |
            |       |
            X-------X

            Height and width are deduced

            rtype: pyscribus.dimensions.DimBox
            """

            tlx,tly = kwargs["top_lx"],kwargs["top_ly"]
            brx,bry = kwargs["bottom_rx"],kwargs["bottom_ry"]

            obj.dims["width"].value = brx - tlx
            obj.dims["height"].value = bry - tly

            obj.setx("top-right", brx)
            obj.sety("top-right", tly)
            obj.setx("top-left", tlx)
            obj.sety("top-left", tly)
            obj.setx("bottom-left", tlx)
            obj.sety("bottom-left", bry)
            obj.setx("bottom-right", brx)
            obj.sety("bottom-right", bry)

            return obj

        def from_tr(obj, kwargs):
            """
            Define DimBox points from top-right corner.

            <--------X
                     |
                     |
                     v

            rtype: pyscribus.dimensions.DimBox
            """

            # Troy to avoid using try and "tory".
            # I don’t care about UK parties.
            trox = float(kwargs["top_rx"])
            troy = float(kwargs["top_ry"])
            width = float(kwargs["width"])
            height = float(kwargs["height"])

            obj.dims["width"].value = width
            obj.dims["height"].value = height

            obj.setx("top-right", trox)
            obj.sety("top-right", troy)
            obj.setx("top-left", trox - width)
            obj.sety("top-left", troy)
            obj.setx("bottom-left", trox - width)
            obj.sety("bottom-left", troy + height)
            obj.setx("bottom-right", trox)
            obj.sety("bottom-right", troy + height)

            return obj

        def from_tl(obj, kwargs):
            """
            Define DimBox points from top-left corner.

            X------->
            |
            |
            v

            rtype: pyscribus.dimensions.DimBox
            """

            tlx = float(kwargs["top_lx"])
            tly = float(kwargs["top_ly"])
            width = float(kwargs["width"])
            height = float(kwargs["height"])

            obj.dims["width"].value = width
            obj.dims["height"].value = height

            obj.setx("top-left", tlx)
            obj.sety("top-left", tly)
            obj.setx("top-right", tlx + width)
            obj.sety("top-right", tly)
            obj.setx("bottom-left", tlx)
            obj.sety("bottom-left", tly + height)
            obj.setx("bottom-right", tlx + width)
            obj.sety("bottom-right", tly + height)

            return obj

        def check_case(kwargs, casename):
            """
            Check if there is any enough kwargs arguments to set the box
            according to a specific box setting scenario / case.

            :rtype: string,boolean
            :returns: casename or False
            """

            case = False

            if casename == "set_from_all":
                case =  ["top_lx", "top_ly", "bottom_rx", "bottom_ry"]

            if casename == "set_from_tl":
                case = ["top_lx", "top_ly", "width", "height"]

            if casename == "set_from_tr":
                case = ["top_rx", "top_ry", "width", "height"]

            if case:

                met = 0

                for k in case:
                    if k in kwargs:
                        met += 1

                if met == len(case):
                    return casename

            return False

        if kwargs is not None:
            case = False

            rotation_deg = False

            if "rotation" in kwargs:
                rotation_deg = kwargs["rotation"]

            # Setting from top left corner ?
            case = check_case(kwargs, "set_from_tl")

            if not case:
                # Setting from top right corner ?
                case = check_case(kwargs, "set_from_tr")

            if not case:
                # Setting from top left and bottom right corner
                case = check_case(kwargs, "set_from_all")

            if case:

                if case == "set_from_tl":
                    self = from_tl(self, kwargs)

                if case == "set_from_tr":
                    self = from_tr(self, kwargs)

                if rotation_deg:
                    self.rotated_coords = copy.deepcopy(self.coords)
                    self.rotate(rotation_deg)

                return True
            else:
                return False
        else:
            return False

    def resize_side(self, side, value):
        """
        Resize the box from a side.

        :type side: str
        :param side: Side of the box to resize from.
            Must be "left", "right", "top", "bottom".
        :type value: float
        :param value: Resize value
        """

        if side in ["left", "right", "top", "bottom"]:

            if side in ["left", "right"]:
                self.dims["width"].value += value

            if side in ["top", "bottom"]:
                self.dims["height"].value += value

            if side == "left":
                nlx = self.getx("top-left") + value
                self.setx("top-left", nlx)
                self.setx("bottom-left", nlx)

            if side == "right":
                nrx = self.getx("top-right") + value
                self.setx("top-right", nrx)
                self.setx("bottom-right", nrx)

            if side == "top":
                nty = self.gety("top-left") + value
                self.sety("top-left", nty)
                self.sety("top-right", nty)

            if side == "bottom":
                nby = self.gety("bottom-left") + value
                self.sety("bottom-left", nby)
                self.sety("bottom-right", nby)

            return True
        else:
            return False

    def rotate(self, degree):
        """
        Rotate the box by degree.

        .. warning:: NOT IMPLEMENTED YET & DON'T KNOW HOW TO

        :type degree: float
        :param degree: Degree of rotation

        This method **don't** modify DimBox.coords but update
        DimBox.rotated_coords.
        """

        # FIXME TODO USE OF MATHEMATICAL BLACK MAGIC REQUIRED HERE
        #
        # Rotate all the points of self.rotated_coords by degree.
        #
        # In case you need to use radians (?) to implement that, please
        # keep in mind that Scribus use degrees as anyone sane and
        # practical do.
        #
        # NOTE It is crucial you don't modify Dim.coords, as in SLA
        # XML, only original box coords and rotation angle value
        # are saved.

        # FIXME Remove that after implementation -------

        logger=logs.getLogger()

        if logger:
            logger.debug("Box rotation not implemented.")
            # logger.debug("Rotation degree : {}".format(degree))
        else:
            print("PyScribus - Box rotation not implemented.")
            # print("Rotation degree :", degree)

        # ----------------------------------------------

        valid = True

        if valid:
            self.rotation.value = degree

        return True

    #--- Python __ methods -----------------------------------------------

    def __eq__(self, other):
        for cname in self.coords.keys():

            for idx in [0,1]:
                a = self.coords[cname][idx].value
                b = other.coords[cname][idx].value

                if not a == b:
                    return False

        return True

    def __str__(self):
        return "X {} Y {} Width {} Height {}".format(
            self.coords["top-left"][0].value,
            self.coords["top-left"][1].value,
            self.dims["width"],
            self.dims["height"],
        )


class LocalDimBox(DimBox):
    """
    Box/rectangle object to manipulate Scribus frames coordinates,
    but coordinates are relative to a parent frame.

    Used for image frames in image objects.

    :type kwargs: dict
    :param kwargs: kwargs (see kwargs table)

    :ivar dict dims: Width and height of the box as Dim objects
    :ivar dict coords: Coordinates of the box, dict of list of
        Dim objects ([x, y]) for each point
    :ivar Dim rotation: Rotation angle of the box as Dim object
        (unit : degree)
    :ivar dict rotated_coords: Coordinates of the box when rotated by
        rotation degree
    :ivar bool visible: Is this box is visible in the parent ?

    +-------------------------+------------+
    | Box point coordinate    | kwargs key |
    +=========================+============+
    | Top left x position     | top_lx     |
    +-------------------------+------------+
    | Top left y position     | top_ly     |
    +-------------------------+------------+
    | Box width (optionnal)   | width      |
    +-------------------------+------------+
    | Box height (optionnal)  | height     |
    +-------------------------+------------+
    """

    def __init__(self, **kwargs):
        DimBox.__init__(self)
        self.visible = True

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Pyscribus submodule for Scribus' headless execution.
"""

# Imports ===============================================================#

import copy
import subprocess

from pathlib import Path

import pyscribus.headless.scripts.topdf as script_to_pdf

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

SCRIPTS = {
    "topdf": script_to_pdf,
}

# Fonctions =============================================================#


def run_pyh_script(script_id: str, **kwargs) -> subprocess.Popen:
    """
    Execute Scribus scripts in command-line.

    :type script_id: str
    :param script_id: ID of the script to run. Must be a
        `pyscribus.headless.SCRIPTS` key.
    :type kwargs: dict
    :param kwargs: Additionnal parameters. See kwargs table.
    :rtype: subprocess.Popen

    +--------------+------------------------------------+
    | Kwargs       | Parameter                          |
    +==============+====================================+
    | sla_input    | Filepath of a SLA file.            |
    +--------------+------------------------------------+
    | stdout       | Output stdout. False by default.   |
    +--------------+------------------------------------+
    | stderr       | Output stderr. False by default.   |
    +--------------+------------------------------------+
    """

    script = SCRIPTS.get(script_id)

    if script is None:
        return False

    base_args = copy.deepcopy(script.SYNTAX)
    current_args = []

    sla_input = kwargs.get("sla_input")

    for arg in base_args:

        if arg == "SCRIPT_PATH":
            current_args.append(script.PATH)
            continue

        if arg == "SLA_INPUT":

            if sla_input is None:
                break

            sla_input = Path(sla_input).resolve(True)

            current_args.append(str(sla_input))
            continue

        current_args.append(arg)

    if not current_args:
        return False

    command = " ".join(current_args)

    if len(current_args) != len(base_args):
        return False

    # stdout & stderr -----------------------------------------------

    get_stdout, get_stderr = kwargs.get("stdout"), kwargs.get("stderr")
    stdout_file, stderr_file = subprocess.DEVNULL, subprocess.DEVNULL

    for std in [[get_stdout, stdout_file], [get_stderr, stderr_file]]:
        if std[0] is not None:
            if std[0]:
                std[1] = subprocess.PIPE

    # ---------------------------------------------------------------

    return subprocess.Popen(
        command,
        shell=True,
        stderr=stderr_file,
        stdout=stdout_file,
    )

# vim:set shiftwidth=4 softtabstop=4:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
ISO 269, international system for paper size, additionnal to ISO 216.

C format, including C7/6 and DL.

See iso216paper module for A, B formats.
See iso217paper module for RA, SRA formats.
"""

# Imports ===============================================================#

from pyscribus.common.math import FloatEnum, PICA_TO_MM

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

class C0(FloatEnum):
    WIDTH = 917 / PICA_TO_MM
    HEIGHT = 1297 / PICA_TO_MM

class C1(FloatEnum):
    WIDTH = 648 / PICA_TO_MM
    HEIGHT = 917 / PICA_TO_MM

class C2(FloatEnum):
    WIDTH = 458 / PICA_TO_MM
    HEIGHT = 648 / PICA_TO_MM

class C3(FloatEnum):
    WIDTH = 324 / PICA_TO_MM
    HEIGHT = 458 / PICA_TO_MM

class C4(FloatEnum):
    WIDTH = 229 / PICA_TO_MM
    HEIGHT = 324 / PICA_TO_MM

class C5(FloatEnum):
    WIDTH = 162 / PICA_TO_MM
    HEIGHT = 229 / PICA_TO_MM

class C6(FloatEnum):
    WIDTH = 114 / PICA_TO_MM
    HEIGHT = 162 / PICA_TO_MM

class C76(FloatEnum):
    WIDTH = 81 / PICA_TO_MM
    HEIGHT = 162 / PICA_TO_MM

class C7(FloatEnum):
    WIDTH = 81 / PICA_TO_MM
    HEIGHT = 114 / PICA_TO_MM

class C8(FloatEnum):
    WIDTH = 57 / PICA_TO_MM
    HEIGHT = 81 / PICA_TO_MM

class C9(FloatEnum):
    WIDTH = 40 / PICA_TO_MM
    HEIGHT = 57 / PICA_TO_MM

class C10(FloatEnum):
    WIDTH = 28 / PICA_TO_MM
    HEIGHT = 40 / PICA_TO_MM

class DL(FloatEnum):
    WIDTH = 110 / PICA_TO_MM
    HEIGHT = 220 / PICA_TO_MM

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
AFNOR (French Standardization Association) certified paper sizes.

"Old" [french] paper sizes.

Provides :

Cloche, Pot / Écolier, Tellière, Couronne écriture, Couronne édition,
Roberto, Écu, Coquille, Carré, Cavalier, Demi-raisin, Raisin,
Double raisin, Jésus, Soleil, Colombier affiche, Colombier commercial,
Petit Aigle, Grand Aigle, Grand Monde, Univers

For more common and international paper sizes, see iso216paper module.
"""

# Imports ===============================================================#

from pyscribus.common.math import FloatEnum, PICA_TO_MM

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

class Cloche(FloatEnum):
    WIDTH = 300 / PICA_TO_MM
    HEIGHT = 400 / PICA_TO_MM

class Pot(FloatEnum):
    """
    Pot paper size. Same as Ecolier.
    """

    WIDTH = 310 / PICA_TO_MM
    HEIGHT = 400 / PICA_TO_MM

class Ecolier(FloatEnum):
    """
    Ecolier paper size. Same as Pot.
    """

    WIDTH = 310 / PICA_TO_MM
    HEIGHT = 400 / PICA_TO_MM

class Telliere(FloatEnum):
    WIDTH = 340 / PICA_TO_MM
    HEIGHT = 440 / PICA_TO_MM

class CouronneEcriture(FloatEnum):
    WIDTH = 360 / PICA_TO_MM
    HEIGHT = 460 / PICA_TO_MM

class CouronneEdition(FloatEnum):
    WIDTH = 370 / PICA_TO_MM
    HEIGHT = 470 / PICA_TO_MM

class Roberto(FloatEnum):
    WIDTH = 390 / PICA_TO_MM
    HEIGHT = 500 / PICA_TO_MM

class Ecu(FloatEnum):
    WIDTH = 400 / PICA_TO_MM
    HEIGHT = 520 / PICA_TO_MM

class Coquille(FloatEnum):
    WIDTH = 440 / PICA_TO_MM
    HEIGHT = 560 / PICA_TO_MM

class Carre(FloatEnum):
    WIDTH = 450 / PICA_TO_MM
    HEIGHT = 560 / PICA_TO_MM

class Cavalier(FloatEnum):
    WIDTH = 460 / PICA_TO_MM
    HEIGHT = 620 / PICA_TO_MM

class Raisin(FloatEnum):
    WIDTH = 500 / PICA_TO_MM
    HEIGHT = 650 / PICA_TO_MM

class DemiRaisin(FloatEnum):
    WIDTH = 325 / PICA_TO_MM
    HEIGHT = 500 / PICA_TO_MM

class DoubleRaisin(FloatEnum):
    WIDTH = 650 / PICA_TO_MM
    HEIGHT = 1000 / PICA_TO_MM

class Jesus(FloatEnum):
    WIDTH = 560 / PICA_TO_MM
    HEIGHT = 720 / PICA_TO_MM

class PetitJesus(FloatEnum):
    WIDTH = 550 / PICA_TO_MM
    HEIGHT = 700 / PICA_TO_MM

class GrandJesus(FloatEnum):
    WIDTH = 560 / PICA_TO_MM
    HEIGHT = 760 / PICA_TO_MM

class Soleil(FloatEnum):
    WIDTH = 600 / PICA_TO_MM
    HEIGHT = 800 / PICA_TO_MM

class ColombierAffiche(FloatEnum):
    WIDTH = 600 / PICA_TO_MM
    HEIGHT = 800 / PICA_TO_MM

class ColombierCommercial(FloatEnum):
    WIDTH = 630 / PICA_TO_MM
    HEIGHT = 900 / PICA_TO_MM

class PetitAigle(FloatEnum):
    WIDTH = 700 / PICA_TO_MM
    HEIGHT = 940 / PICA_TO_MM

class GrandAigle(FloatEnum):
    WIDTH = 750 / PICA_TO_MM
    HEIGHT = 1060 / PICA_TO_MM

class GrandMonde(FloatEnum):
    WIDTH = 900 / PICA_TO_MM
    HEIGHT = 1260 / PICA_TO_MM

class Univers(FloatEnum):
    WIDTH = 1000 / PICA_TO_MM
    HEIGHT = 1300 / PICA_TO_MM

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus classes for patterns.
"""

# Imports ===============================================================#

import lxml
import lxml.etree as ET

import pyscribus.exceptions as exceptions
import pyscribus.dimensions as dimensions
import pyscribus.pageobjects as pageobjects

from pyscribus.common.xml import *

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

class Pattern(PyScribusElement):
    """
    Pattern in SLA (Pattern)

    :type sla_parent: pyscribus.sla.SLA
    :param sla_parent: SLA parent instance
    :type doc_parent: pyscribus.document.Document
    :param doc_parent: SLA DOCUMENT instance
    """

    # <Pattern
    #     xoffset="0"
    #     yoffset="0"
    # >

    def __init__(self, sla_parent=False, doc_parent=False):
        super().__init__()

        self.name = None

        self.items = []

        self.sla_parent = sla_parent
        self.doc_parent = doc_parent

        # NOTE We do not set a dimensions.DimBox because patterns is not meant
        # to be used directly on the page

        self.dims = {
            "width": None,
            "height": None
        }

        self.scale = {
            "x": None,
            "y": None
        }

    def fromxml(self, xml):
        """
        """

        if xml.tag == "Pattern":

            name = xml.get("Name")
            if name is not None:
                self.name = name

            for dim in ["width", "height"]:

                att = xml.get(dim)
                if att is not None:
                    self.dims[dim] = dimensions.Dim(float(att))

            for scale in ["x", "x"]:
                att_name = "scale{}".format(scale.upper())

                att = xml.get(att_name)
                if att is not None:
                    self.scale[scale] = dimensions.Dim(float(att), unit="pcdecim")

            # TODO FIXME xoffset yoffset

            for element in xml:

                if element.tag == "PatternItem":
                    pie = PatternItem(self.sla_parent, self.doc_parent)

                    success = pie.fromxml(element)
                    if success:
                        self.items.append(pie)

            return True

        return False

    def toxml(self):
        """
        """

        xml = ET.Element("Pattern")

        if self.name is not None:
            xml.attrib["Name"] = self.name

        for dim in ["width", "height"]:
            if self.dims[dim] is None:
                raise ValueError(
                    "Pattern {} must have a(n) {}.".format(
                        self.name, dim
                    )
                )
            else:
                xml.attrib[dim] = self.dims[dim].toxmlstr()

        for scale in ["x", "y"]:
            att_name = "scale{}".format(scale.upper())

            if self.scale[scale] is None:
                xml.attrib[att_name] = "1"
            else:
                xml.attrib[att_name] = self.scale[scale].toxmlstr()

        # TODO FIXME xoffset yoffset

        for item in self.items:
            ix = item.toxml()
            xml.append(ix)

        return xml


class PatternItem(pageobjects.PageObject):
    """
    Pattern item in SLA (Pattern/PatternItem)

    (Pattern items are encoded as polygon page objects with a different XML tag)

    :type sla_parent: pyscribus.sla.SLA
    :param sla_parent: SLA parent instance
    :type doc_parent: pyscribus.document.Document
    :param doc_parent: SLA DOCUMENT instance
    """

    # <PatternItem XPOS="0.187499995780351" YPOS="0.187499995780351" OwnPage="-1" ItemID="204434288" PTYPE="6" WIDTH="6.55552657556513" HEIGHT="4.82126570393049" FRTYPE="3" CLIPEDIT="1" PWIDTH="0.374999991560701" PCOLOR="FromSVG#de4344" PCOLOR2="FromSVG#de4344" PLINEART="1" PLINEJOIN="128" ANNAME=" path861" TEXTFLOWMODE="1" path="M3.6e-7 4.44089e-16 L6.55553 2.41063 L0 4.82127 C1.0473 3.39803 1.04126 1.45079 3.6e-7 0 L3.6e-7 4.44089e-16 Z" copath="M3.6e-7 4.44089e-16 L6.55553 2.41063 L0 4.82127 C1.0473 3.39803 1.04126 1.45079 3.6e-7 0 L3.6e-7 4.44089e-16 Z" gXpos="0.187499995780351" gYpos="0.187499995780351" gWidth="6.93052656712583" gHeight="5.19626569549119" LAYER="0"/>

    def __init__(self, sla_parent=False, doc_parent=False, **kwargs):
        pageobjects.PageObject.__init__(self, "polygon", sla_parent, doc_parent)
        pageobjects.PageObject._quick_setup(self, kwargs)

    def fromxml(self, xml):
        """
        """

        success = pageobjects.PageObject.fromxml(
            self, xml, arbitrary_tag="PatternItem"
        )

        return success

    def toxml(self):
        """
        """

        xml = pageobjects.PageObject.toxml(arbitrary_tag="PatternItem")

        return xml

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus classes for notes
"""

# Imports ===============================================================#

import lxml
import lxml.etree as ET

import pyscribus.exceptions as exceptions

from pyscribus.common.xml import *

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

class Note(PyScribusElement):
    """
    Note element (Note)
    """

    def __init__(self):
        PyScribusElement.__init__(self)

        self.style = ""
        self.parent_mark = ""
        self.parent_frame = ""
        self.text = NoteText()

    def fromdefault(self):
        self.style = "Default"
        self.text = NoteText()
        self.text.fromdefault()

    def get_parent_mark(self):
        return "NoteMark_{} in frame {}".format(
            self.style, self.parent_frame
        )

    def fromxml(self, xml):

        if xml.tag == "Note":
            style = xml.get("NStyle")
            parent_mark = xml.get("Master")

            if style is not None:
                self.style = style

            if parent_mark is not None:
                self.parent_mark = parent_mark
                self.parent_frame = self.parent_mark.split("in frame ")[-1]

            if "Text" in xml.attrib:
                text = NoteText()
                success = text.fromxml(xml)

                if success:
                    self.text = text

            return True
        else:
            return False

    def toxml(self):
        xml = ET.Element("Note")

        xml.attrib["Master"] = self.get_parent_mark()
        xml.attrib["NStyle"] = self.style
        xml.attrib["Text"] = self.text.toxmlstr()

        return xml


class NoteText(PyScribusElement):
    """
    Note element text (Note/@Text).

    Note element text consists in a full XML file dumped into
    @Text attribute of Note, hence the need for a specific
    class.
    """

    #Text="&lt;?xml version=&quot;1.0&quot; encoding=&quot;UTF-8&quot;?&gt;&lt;SCRIBUSTEXT &gt;&lt;defaultstyle /&gt;&lt;p &gt;&lt;style /&gt;&lt;span &gt;&lt;charstyle Features=&quot;inherit &quot; /&gt; Pri tiuj ĉi ideoj « estas modo » paroli ne alie, ol kun ironia kaj malestima rideto, tial tiel agas ankaŭ A kaj B kaj C, kaj ĉiu el ili timas enpensiĝi serioze eĉ unu minuton pri la mokata ideo, ĉar li « scias antaŭe », ke « ĝi krom malsaĝaĵo enhavas ja nenion », kaj li timas, ke oni iel alkalkulos lin mem al la nombro de « tiuj malsaĝuloj », se li eĉ en la daŭro de unu minuto provos rilati serioze al tiu ĉi malsaĝaĵo. La homoj miras, « kiamaniere en nia praktika tempo povas aperi tiaj malsaĝaj fantaziuloj kaj kial oni ne metas ilin en la domojn por frenezuloj ».&lt;/span&gt;&lt;/p&gt;&lt;/SCRIBUSTEXT&gt;&#10;"/>

    #Text="
    # <?xml version="1.0" encoding="UTF-8"?>
    # <SCRIBUSTEXT >
        # <defaultstyle />
        # <p >
            # <style />
            # <span >
                # <charstyle Features="inherit " />
                # Pri tiuj ĉi ideoj « estas modo » paroli ne alie, ol kun ironia kaj malestima rideto, tial tiel agas ankaŭ A kaj B kaj C, kaj ĉiu el ili timas enpensiĝi serioze eĉ unu minuton pri la mokata ideo, ĉar li « scias antaŭe », ke « ĝi krom malsaĝaĵo enhavas ja nenion », kaj li timas, ke oni iel alkalkulos lin mem al la nombro de « tiuj malsaĝuloj », se li eĉ en la daŭro de unu minuto provos rilati serioze al tiu ĉi malsaĝaĵo. La homoj miras, « kiamaniere en nia praktika tempo povas aperi tiaj malsaĝaj fantaziuloj kaj kial oni ne metas ilin en la domojn por frenezuloj ».
            # </span>
        # </p>
    # </SCRIBUSTEXT>
    # &#10;"/>

    def __init__(self):
        PyScribusElement.__init__(self)

    def fromxml(self, xml):

        if xml == "Note":
            content = xml.get("Text")

            if content is not None:
                text = content

                # &#10; = Line feed character

                for normalise in [
                            [" &gt;&lt;","><"], ["&lt;", "<"], ["&gt;", ">"],
                            ["&quot;", '"'], ["&#10;",""]
                        ]:
                    text = text.replace(normalise[0], normalise[1])

                xml_text = ET.fromstring(text)

                return True
            else:
                return False
        else:
            return False

    def toxml(self):
        """
        Alias of toxmlstr()
        """
        return self.toxmlstr()

    def toxmlstr(self):
        xml_string = ""

        # TODO

        return xml_string


class NoteFrame(PyScribusElement):
    """
    Note frame (NotesFrames/FOOTNOTEFRAME).
    """

    def __init__(self):
        PyScribusElement.__init__(self)

        # NOTE @myID
        # The page object where the note content is located
        self.own_frame_id = None

        # NOTE @MasterID
        # The page object with the story where note references are located
        self.story_frame_id = None

        # NOTE @NSname
        self.note_style = None

    def fromxml(self, xml):

        if xml == "FOOTNOTEFRAME":
            note_style = xml.get("NSname")
            own_frame = xml.get("myID")
            story_frame = xml.get("MasterID")

            if note_style is not None:
                self.note_style = note_style

            if own_frame is not None:
                self.own_frame_id = own_frame

            if story_frame is not None:
                self.story_frame_id = story_frame

            if story_frame is None or own_frame is None:

                raise exceptions.InsaneSLAValue(
                    "Note frame has no page object ID or\
                     no parent page object ID."
                )

            return True

        return False

    def toxml(self):
        xml = ET.Element("FOOTNOTEFRAME")

        if self.note_style is not None:
            xml.attrib["NSname"] = self.note_style

        if self.own_frame_id is not None:
            xml.attrib["myID"] = self.own_frame_id

        if self.story_frame_id is not None:
            xml.attrib["MasterID"] = self.story_frame_id

        return xml

    def fromdefault(self):
        self.note_style = "Default"

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus submodules that may require additional python modules.
"""

__author__ = "Etienne Nadji <etnadji@eml.cc>"
__version__ = "0.1"

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Item attributes for document, page objects.
"""

# Imports ===============================================================#

import lxml
import lxml.etree as ET

import pyscribus.exceptions as exceptions

from pyscribus.common.xml import *

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

class ItemAttribute(PyScribusElement):
    attrib_types = ["boolean", "integer", "double", "string", "none"]

    #--- FIXME Not documented -----------------------------
    # Parameter=""
    # Relationship="none"
    # RelationshipTo=""
    # AutoAddTo="none"

    #--- NOTE Examples in SLA -----------------------------
    # Name="entier" Type="integer" Value="1"
    # Name="reel" Type="double" Value="-5"
    # Name="chaine" Type="string" Value="&quot;Test&quot;"
    # Name="" Type="none" Value=""

    def __init__(self):
        super().__init__()

        self.name = ""
        self.attribute_type = "none"
        self.value = ""

    def as_python(
            self, bool_keywords={"true": "True", "false": "False"},
            lower_bool=False):
        """
        Returns the value of ItemAttribute.value as a valid python
        object (int for integer, bool for boolean, etc).

        :type bool_keywords: dict
        :param bool_keywords: Dictionnary of true and false values to test
            against. If bool_keywords["true"] is equal to
            ItemAttribute.value (as a string), True is returned, otherwise
            False is returned. Example : {"true": "1", "false": "0"}
        :type lower_bool: boolean
        :param lower_bool: In the case of boolean ItemAttribute type, it may
            be useful to switch ItemAttribute.value as a lowercase string to
            avoid case issues with bool_keywords.

        Returns :

        - int object if attribute type is "integer"
        - float object if attribute type is "double"
        - str object if attribute type is "string"
        - bool object if attribute type is "boolean"
        - str if attribute_type is "none", as attribute_type "none" allows
            other values than None.

        Raise a ValueError if attribute_type is empty/unknown.

        :rtype: int,long,string,boolean
        """

        if self.attribute_type == "integer":
            return int(self.value)
        elif self.attribute_type == "double":
            return float(self.value)
        elif self.attribute_type == "string":
            # FIXME TODO Escape &quot; and other entities in unicode caracters
            return str(self.value)
        elif self.attribute_type == "none":
            return self.value
        elif self.attribute_type == "boolean":
            if lower_bool:
                v = str(self.value).lower()
            else:
                v = str(self.value)

            if v == bool_keywords["true"]:
                return True
            else:
                return False
        else:
            raise exceptions.UnknownOrEmptyItemAttributeType()

    #--- PyScribus standard methods ----------------------------

    def fromxml(self, xml):
        name = xml.get("Name")

        if name is not None:
            self.name = name

        atype = xml.get("Type")

        if atype is not None:
            if atype.lower() in ItemAttribute.attrib_types:
                self.attribute_type = atype.lower()

        v = xml.get("Value")

        if v is not None:
            self.value = v

        # TODO

        return True

    def toxml(self):
        xml = ET.Element("ItemAttribute")

        xml.attrib["Name"] = self.name
        xml.attrib["Type"] = self.attribute_type
        xml.attrib["Value"] = self.value

        # TODO

        return xml

    def fromdefault(self):
        # TODO

        self.attribute_type = "none"


class DocumentAttribute(ItemAttribute):
    """
    Item attribute at document level
    """

    def __init__(self):
        super().__init__()


class PageObjectAttribute(ItemAttribute):
    """
    Item attribute at page object level
    """

    def __init__(self):
        super().__init__()

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Classes related to marks management
"""

# Imports ===============================================================#

import lxml
import lxml.etree as ET

import pyscribus.common.xml as xmlc
import pyscribus.exceptions as exceptions

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

mark_type_xml = {
    "anchor": "0",
    "objectref": "1",
    "markref": "2",
    "variable": "3",
    "note": "4",
}

# Classes ===============================================================#

class DocumentMark(xmlc.PyScribusElement):
    """
    Mark element (DOCUMENT/Marks/Mark)
    """

    def __init__(self):
        global mark_type_xml

        super().__init__()

        self.type = ""
        self.name = ""
        self.label = ""

        self.target_object = None
        self.target_mark = {"label": None, "type": "-1"}

        self.pyscribus_defaults = [k for k in mark_type_xml.keys()]

    def fromdefault(self, default):
        """
        :type default: string
        :param default: Set of default settings to apply
        :rtype: boolean
        """

        if default in self.pyscribus_defaults:
            self.type = default

            return True
        else:
            return False

    def set_type(self, mtype):
        """
        :type mtype: string
        :param mtype: Mark type in pyscribus.marks.mark_type_xml keys()
        :rtype: boolean
        :returns: True if setting type succeed
        """
        global mark_type_xml

        if mtype in mark_type_xml.keys():
            self.type = mark_type_xml[mtype]
            return True

        else:
            return False

    def toxml(self):
        """
        :rtype: lxml.etree._Element
        :returns: XML representation of document mark
        """

        global mark_type_xml

        xml = ET.Element("Mark")

        if self.type:

            # --- Label -----------------------------------------------------

            # When @type is "3" (variable text), @str acts as @label
            # in other @types, and @label acts as a mark identifier.
            #
            # So :
            #   DocumentMark.name  = @label if @type == "3"
            #   DocumentMark.label = @label if @type != "3"

            if self.type == "variable":
                xml.attrib["label"] = self.name
            else:
                xml.attrib["label"] = self.label

            # ---------------------------------------------------------------

            xml.attrib["type"] = mark_type_xml[self.type]

            if self.type == "variable":
                xml.attrib["str"] = self.label

            if self.type == "objectref":
                xml.attrib["ItemID"] = self.target_object

            if self.type == "markref":
                xml.attrib["MARKlabel"] = self.target_mark["label"]
                xml.attrib["MARKtype"] = mark_type_xml[self.target_mark["type"]]

        else:
            raise exceptions.InsaneSLAValue("Invalid Marks/Mark @type")

        return xml

    def fromxml(self, xml):
        """
        :rtype: boolean
        :returns: True if XML parsing succeed
        """

        global mark_type_xml

        if xml.tag == "Mark":
            mtype = xml.get("type")

            if mtype is not None:
                for h,x in mark_type_xml.items():
                    if mtype == x:
                        self.type = h
                        break

            # --- Name and/or label -----------------------------------------

            # When @type is "3" (variable text), @str acts as @label
            # in other @types, and @label acts as a mark identifier.
            #
            # So :
            #   DocumentMark.name  = @label if @type == "3"
            #   DocumentMark.label = @label if @type != "3"

            mlabel = xml.get("label")
            if mlabel is not None:
                if self.type == "variable":
                    self.name = mlabel
                else:
                    self.label = mlabel

            if self.type == "variable":
                mstr = xml.get("str")
                if mstr is not None:
                    self.label = mstr

            # ---------------------------------------------------------------

            if self.type == "objectref":
                mitem = xml.get("ItemID")
                if mitem is not None:
                    self.target_object = mitem

            if self.type == "markref":
                mtarget_label = xml.get("MARKlabel")
                if mtarget_label is not None:
                    self.target_mark["label"] = mtarget_label

                mtarget_type = xml.get("MARKtype")
                if mtarget_type is not None:

                    if mtarget_type in mark_type_xml.values():
                        self.target_mark["type"] = mark_type_xml[mtarget_type]
                    else:
                        raise exceptions.InsaneSLAValue(
                            "Invalid Marks/Mark @type"
                        )

            return True

        else:
            return False


class StoryMarkAbstract(xmlc.PyScribusElement):
    """
    Abstract class for MARK elements in Scribus stories.

    :type mark_type: str
    :param mark_type: Type of mark
    :type label: str
    :param label: Mark label
    :type features: dict
    :param features: Text formatting features as dict

    .. seealso:: :class:`pyscribus.stories.StoryNoteMark`
    """

    def __init__(self, mark_type, label="", features=False):
        global mark_type_xml

        super().__init__()

        self.features = {
            "inherit": False,
            "superscript": False,
        }

        self.label = label

        if mark_type in mark_type_xml:
            self.type = mark_type
        else:
            raise ValueError("Unknown mark type")

        if features:
            self.set_features(features)

    def fromxml(self, xml):
        global mark_type_xml

        if xml.tag == "MARK":

            mtype = xml.get("type")
            if mtype is not None:
                for h,x in mark_type_xml.items():
                    if mtype == x:
                        self.type = h
                        break

            mlabel = xml.get("label")
            if mlabel is not None:
                self.label = mlabel

            mfeatures = xml.get("features")
            if mfeatures is not None:
                self.set_features(mfeatures)

            return True

        return False

    def toxml(self):
        global mark_type_xml

        xml = ET.Element("MARK")

        xml.attrib["type"] = mark_type_xml[self.type]
        xml.attrib["label"] = self.label

        have_features = len([f for f in self.features.values() if f])

        if have_features:
            features = []

            for f,v in self.features.items():
                if v:
                    features.append(f)

            features = " ".join(features)

            xml.attrib["FEATURES"] = features

        return xml

    def set_features(self, features):
        """
        :type features: str
        :param features: Formatting features as string separated by spaces.
            Ex: ``"inherit superscript"``
        :rtype: bool
        :returns: False if one features was not set.
        """

        features = features.split()

        fset = 0

        for feature in features:
            if feature in self.features.keys():
                self.features[feature] = True
                fset += 1

        if fset == len(features):
            return True
        else:
            return False


class StoryNoteMark(StoryMarkAbstract):
    """
    Mark (MARK) for a (foot|end)note in Scribus stories.
    """

    def __init__(self, label="", features=False):
        StoryMarkAbstract.__init__(self, "note", label, features)

    def fromdefault(self):
        self.set_features("inherit superscript")

    def __repr__(self):
        return "NOTEMARK|{}|{}".format(
            self.label,
            [k for k,v in self.features.items() if self.features[k]]
        )


class StoryVariableMark(StoryMarkAbstract):
    """
    Variable text mark
    """

    def __init__(self, label=""):
        StoryMarkAbstract.__init__(self, "variable", label, False)


class StoryAnchorMark(StoryMarkAbstract):
    """
    Anchor mark
    """

    def __init__(self, label=""):
        StoryMarkAbstract.__init__(self, "anchor", label, False)


class StoryPageObjectRefMark(StoryMarkAbstract):
    """
    Reference to a page object mark
    """

    def __init__(self, label=""):
        StoryMarkAbstract.__init__(self, "objectref", label, False)


class StoryMarkRefMark(StoryMarkAbstract):
    """
    Reference to a mark... mark
    """

    def __init__(self, label=""):
        StoryMarkAbstract.__init__(self, "markref", label, False)

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
SLA file.
"""

# Imports ===============================================================#

import re

import lxml
import lxml.etree as ET

import pyscribus.common.xml as xmlc

import pyscribus.exceptions as exceptions
import pyscribus.dimensions as dimensions
import pyscribus.document as document
import pyscribus.stories as stories
import pyscribus.pageobjects as pageobjects

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

class SLA(xmlc.PyScribusElement):
    """
    SLA file.

    :type filepath: str
    :param filepath: SLA file path
    :type version: str
    :param version: Scribus version (ex. '1.5.1')
    :type kwargs: dict
    :param kwargs: kwargs

    **Kwargs options :**

    +-----------------------+---------------------------+---------------+
    | Kwarg key             | Use                       | Default value |
    +=======================+===========================+===============+
    | templating            | Use templating ?          | False         |
    +-----------------------+---------------------------+---------------+
    | templatingInsensitive | Should in-text templating | False         |
    |                       | be case insensitive ?     |               |
    +-----------------------+---------------------------+---------------+
    | templatingPattern     | compiled regex to find    | \^%\w+%$      |
    |                       | templated elements        |               |
    |                       | (ex: %TITLE%)             |               |
    +-----------------------+---------------------------+---------------+
    """

    def __init__(self, filepath="", version="", **kwargs):
        super().__init__()

        if version:
            self.version = version.split(".")

        self.document = None

        self.templating = {
            "active": False,
            # In text templating sequences are like %Title%
            "intext-pattern": re.compile("(%\w+%)+"),
            # %Title% = %title% = %TITLE% ?
            "intext-insensitive": False,
            # Page object attribute templating names are like %Title%
            "attribute-pattern": re.compile("(%\w+%)+"),
            # dict for evaluating an item attribute value as python boolean
            "attribute-eval-bool-keywords": {"true": "True", "false": "False"},
            # Insensitive item attribute value evaluation as boolean ?
            "attribute-eval-bool-insensive": False
        }

        for argname, argvalue in kwargs.items():

            if argname == "templating":
                if argvalue:
                    self.templating["active"] = True

            if argname == "templatingInsensitive":
                if argvalue:
                    self.templating["intext-insensitive"] = True

            if argname == "templatingPattern":

                try:
                    pattern = re.compile(argvalue)
                    self.templating["intext-pattern"] = pattern

                except TypeError:

                    raise TypeError(
                        "templating-pattern must be a re pattern string."
                    )


        if filepath:
            self.parse(filepath, kwargs)

    def fromdefault(self):
        """
        Set default attributes.

        Add a default document.
        """

        doc = document.Document()
        doc.fromdefault()
        self.append(doc)

        return True

    def append(self, sla_object):
        """
        Add the document to SLA Documents and set its sla_parent as self

        :type sla_object: pyscribus.document.Document, other
        :param sla_object: Document or any PyScribus object appendable to a document.
        :type document_index: int
        :param document_index: Index of document if sla_object is not one
        :rtype: boolean
        :returns: True if appending succeed
        """

        if isinstance(sla_object, document.Document):
            sla_object.sla_parent = self
            self.document = sla_object

            return True
        else:
            if self.document is None:
                return False
            else:
                return self.document.append(sla_object)

    def save(self, filepath):
        """
        Save SLA file.

        :type filepath: str
        :param filepath: SLA file path
        :rtype: boolean
        :returns: True if successfull
        """

        xml = self.toxml()

        xml_string = '<?xml version="1.0" encoding="UTF-8"?>' + "\n"
        xml_string += str(
            ET.tostring(
                xml,
                encoding="unicode",
                pretty_print=True
            )
        )

        with open(filepath, "w", encoding="utf8") as slaf:
            slaf.write(xml_string)

    def toxml(self, optional=True):
        """
        Return SLA as lxml.etree._Element

        :type optional: bool
        :param optional: Includes optional attributes (True by default)
        :returns: xml
        :rtype: lxml._Element
        """

        xml = ET.Element("SCRIBUSUTF8NEW")

        xml.attrib["Version"] = ".".join(self.version)

        if self.document is None:
            raise exceptions.InsaneSLAValue(
                "SLA file has no SCRIBUSUTF8NEW/DOCUMENT"
            )
        else:
            dx = self.document.toxml(optional)
            xml.append(dx)

        return xml

    def stories(self, templatable=False):
        """
        Returns all stories from text frames (TextObject) in SLA.

        :param templatable: Only return stories in which there is templatable
            content
        :type templatable: boolean
        :returns: List of pyscribus.story.Story
        :rtype: list
        """

        all_stories = self.document.stories()

        if templatable and self.templating["active"]:
            return [story for story in all_stories if story.templatable()]

        return stories

    def pageobjects(self, object_type=False, templatable=False):
        """
        """

        return self.document.pageobjects(object_type, templatable)

    def templatable_stories(self):
        """
        Returns stories from text frames (TextObject) in SLA which contains
        templatable content.
        """

        return self.stories(templatable=True)

    def parse(self, filepath, kwargs):
        """
        Import SLA data from a file path.

        :type filepath: str
        :param filepath: SLA file path
        :type kwargs: dict
        :param kwargs: kwargs
        :returns: True if successfull parsing
        :rtype: boolean
        """

        xml = ET.parse(filepath).getroot()
        success = self.fromxml(xml)

        return success

    def fromxml(self, xml):
        """
        Set SLA content according to <xml> (lxml.etree._Element).

        Use SLA.parse() if you want to import from a file path.

        :returns: True if successfull parsing
        :rtype: boolean

        .. seealso:: pyscribus.sla.SLA.parse()
        """

        def read_xml(obj, xml):
            if xml.tag == "SCRIBUSUTF8NEW":

                version = xml.get("Version")
                if version is not None:
                    obj.version = version.split(".")

                for element in xml:

                    if element.tag == "DOCUMENT":
                        doc = document.Document(sla_parent=self)
                        success = doc.fromxml(element)

                        if success:
                            obj.document = doc

                return True, obj

            else:
                return False, obj

        success = False

        if isinstance(xml, lxml.etree._Element):
            success, self = read_xml(self, xml)
        else:
            if os.path.exists(os.path.realpath(xml)):
                xml = lxml.parse(filepath).getroot()
                success, self = read_xml(self, xml)

            else:
                raise TypeError("fromxml requires lxml.etree._Element.")

        return success

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
PyScribus classes for colors and gradients.
"""

# Imports ===============================================================#

import lxml
import lxml.etree as ET

import pyscribus.common.xml as xmlc
import pyscribus.common.math as pmath
import pyscribus.exceptions as exceptions
import pyscribus.dimensions as dimensions

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

class Color(xmlc.PyScribusElement):
    """
    SLA Color (COLOR)

    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    +------------+---------------------------------+-----------+
    | Kwargs     | Setting                         | Type      |
    +============+=================================+===========+
    | default    | Equivalent to a ``fromdefault`` | boolean   |
    |            | call.                           | or string |
    |            |                                 |           |
    |            | Value being True or the name of |           |
    |            | the default set.                |           |
    +------------+---------------------------------+-----------+
    """

    defaults = {
        "Black": {
            "space": "cmyk",
            "colors": [0, 0, 0, 100]
        },
        "Blue": {
            "space": "rgb",
            "colors": [0, 0, 255]
        },
        "Cool Black": {
            "space": "cmyk",
            "colors": [60, 0, 0, 100]
        },
        "Cyan": {
            "space": "cmyk",
            "colors": [100, 0, 0, 0]
        },
        "Green": {
            "space": "rgb",
            "colors": [0, 255, 0]
        },
        "Magenta": {
            "space": "cmyk",
            "colors": [0, 100, 0, 0]
        },
        "Red": {
            "space": "rgb",
            "colors": [255, 0, 0]
        },
        "Registration": {
            "space": "cmyk",
            "colors": [100, 100, 100, 100],
            "register": True
        },
        "Rich Black": {
            "space": "cmyk",
            "colors": [60, 40, 40, 100]
        },
        "Warm Black": {
            "space": "cmyk",
            "colors": [0, 60, 29.8039215686275, 100]
        },
        "White": {
            "space": "cmyk",
            "colors": [0, 0, 0, 0]
        },
        "Yellow": {
            "space": "cmyk",
            "colors": [0, 0, 100, 0]
        },
    }

    def __init__(
            self, name="Black", space="CMYK", colors=[0,0,0,100],
            register="0", **kwargs):
        super().__init__()

        self.pyscribus_defaults = [k for k in Color.defaults.keys()]

        self.name = name
        self.register = register
        self.set_space_colors(space, colors)

        self._quick_setup(kwargs)

    #--- Color management --------------------------------------

    def set_space_colors(self, space, colors):
        """
        Set color space and color inks of the colors.

        :type space: string
        :param space: Color space. Either "cmyk" or "rgb".
        :type colors: list
        :param colors: List of inks values (float).

        :Example:

        .. code:: python

           # Green RGB
           color.set_space_colors("rgb", [0, 255, 0])

           # White CMYK
           color.set_space_colors("cmyk", [0, 0, 0, 0])

        """

        spaced = self.set_space(space)

        if spaced:
            self.set_colors(colors)

    def set_colors(self, colors, space=False):

        if not space:
            if self.is_cmyk:
                space = "cmyk"
            else:
                space = "rgb"

        if space.lower() in ["cmyk", "rgb"]:

            if space.lower() == "cmyk":
                self.colors = {
                    "C": float(colors[0]),
                    "M": float(colors[1]),
                    "Y": float(colors[2]),
                    "K": float(colors[3])
                }

                # Avoid invalid values

                for ink in ["C", "M", "Y", "K"]:
                    if self.colors[ink] > float(100.0):
                        self.colors[ink] = 100.0

                for ink in ["C", "M", "Y", "K"]:
                    if self.colors[ink] < float(0.0):
                        self.colors[ink] = 0.0

            if space.lower() == "rgb":
                self.colors = {
                    "R": float(colors[0]),
                    "G": float(colors[1]),
                    "B": float(colors[2])
                }

                # Avoid invalid values

                for ink in ["R", "G", "B"]:
                    if self.colors[ink] > float(255.0):
                        self.colors[ink] = 255.0

                for ink in ["R", "G", "B"]:
                    if self.colors[ink] < float(0.0):
                        self.colors[ink] = 0.0

            return True

        return False

    def set_space(self, space):
        """
        Set the color space (CMYK / RGB) of the color.

        :type space: string
        :param space: Color space. Either "cmyk" or "rgb".
        """

        if space.lower() in ["cmyk", "rgb"]:
            if space.lower() == "cmyk":
                self.is_cmyk = True
                self.is_rvb = False

            if space.lower() == "rgb":
                self.is_cmyk = False
                self.is_rvb = True

            return True

        return False

    #--- PyScribus standard methods ----------------------------

    def fromdefault(self, default="Black"):
        """
        Set default attributes for this color.

        :Example:

        .. code:: python

           # with fromdefault method :
           black = Color()
           black.fromdefault("Black")

           # with quick setup :
           blue = Color(default="Blue")

        .. seealso::
            **For fromdefault explanation**:
            `pyscribus.common.xml.PyScribusElement.fromdefault`

            **For current class default sets**:
            `pyscribus.colors.Color.listdefaults`

        """

        if default in self.pyscribus_defaults:

            self.name = default

            self.set_space_colors(
                Color.defaults[default]["space"],
                Color.defaults[default]["colors"]
            )

            if "register" in Color.defaults[default]:
                self.register = True

            return True

        return False

    def fromxml(self, xml):
        """
        :type xml: lxml._Element
        :param xml: XML source of color
        :rtype: boolean
        :returns: True if XML parsing succeed
        """

        space = xml.get("SPACE")

        if space is not None:
            # Name -------------------------------------------------------

            name = xml.get("NAME")

            if name is not None:
                self.name = name

            # Space ------------------------------------------------------

            if space.lower() in ["cmyk", "rgb"]:
                self.set_space(space)

                # Colors -------------------------------------------------

                if self.is_cmyk:
                    colors = [xml.get(c) for c in ["C", "M", "Y", "K"]]
                else:
                    colors = [xml.get(c) for c in ["R", "G", "B"]]

                if colors:
                    if None in colors:
                        raise exceptions.InvalidColor(
                            "Invalid inks <{}>.".format(",".join(colors))
                        )
                    else:
                        self.set_colors(colors)
                else:
                    raise exceptions.InvalidColor("No inks.")

            # Register ---------------------------------------------------

            reg = xml.get("Register")

            if reg is None:
                self.register = False
            else:
                if int(reg):
                    self.register = True
                else:
                    self.register = False

            # ------------------------------------------------------------

            return True

        else:
            return False

    def toxml(self):
        """
        :rtype: lxml._Element
        :returns: Color as XML element
        """

        xml = ET.Element("COLOR")
        xml.attrib["NAME"] = self.name

        if self.is_cmyk:
            xml.attrib["SPACE"] = "CMYK"

            for color in ["C", "M", "Y", "K"]:
                color_value = pmath.necessary_float(self.colors[color])
                xml.attrib[color] = str(color_value)

        else:
            xml.attrib["SPACE"] = "RGB"

            for color in ["R", "G", "B"]:
                color_value = pmath.necessary_float(self.colors[color])
                xml.attrib[color] = str(color_value)

        if self.register:
            xml.attrib["Register"] = "1"

        return xml

    def _quick_setup(self, settings: dict):
        """
        Method for defining gradient stop settings from class
        instanciation kwargs.

        :type settings: dict
        :param settings: Kwargs dictionnary
        """

        if settings:
            xmlc.PyScribusElement._quick_setup(self, settings)

            for setting_name, setting_value in settings.items():
                pass

    #--- Python __ methods -------------------------------------

    def __eq__(self, other):
        """
        Equality operator.

        Doesn’t check if RVB color can be an equivalent in CMYK.
        """

        # NOTE Obviously we don't translate colors of different spaces
        # to compare their inks, as RVB and CMYK color spectrums are
        # not the same.

        same_space = False

        if self.is_cmyk and other.is_cmyk:
            same_space = True

        if self.is_rvb and other.is_rvb:
            same_space = True

        if same_space:
            # NOTE Self colors and other colors are in the same space, so
            # inks dicts have the same keys. So we get a list of
            # [(R1,R2)…] or [(C1,C2)…]. Having one ink different is enough.
            inks = zip(self.colors.values(), other.values())

            for ink in inks:
                if ink[0] != ink[1]:
                    return False
        else:
            return False

        return True

    def __repr__(self):
        if self.is_cmyk:
            s = "CMYK"
            inks = ["C", "M", "Y", "K"]
        else:
            s = "RGB"
            inks = ["R", "G", "B"]

        r = "{}:{}:{}:{}".format(
            self.name, s,
            ";".join(
                [
                    str(self.colors[ink])
                    for ink in inks
                ]
            ),
            self.register
        )

        return r


class GradientColorStop(xmlc.PyScribusElement):
    """
    Gradient color stop (Gradient/CSTOP)

    :type kwargs: dict
    :param kwargs: Quick setting (see kwargs table)

    +------------+---------------------------------+-----------+
    | Kwargs     | Setting                         | Type      |
    +============+=================================+===========+
    | default    | Equivalent to a ``fromdefault`` | boolean   |
    |            | call.                           | or string |
    |            |                                 |           |
    |            | Value being True or the name of |           |
    |            | the default set.                |           |
    +------------+---------------------------------+-----------+
    | opacity    | Percentage of opacity,          | float     |
    |            | from 0 to 1.                    |           |
    +------------+---------------------------------+-----------+
    | color      | Color name                      | string    |
    +------------+---------------------------------+-----------+
    | position   |                                 |           |
    +------------+---------------------------------+-----------+
    | shade      |                                 |           |
    +------------+---------------------------------+-----------+

    :Example:

    .. code:: python

       cstop1 = colors.GradientColorStop(
           color="Black", shade=100, position=0, opacity=1
       )

    """

    def __init__(self, **kwargs):
        super().__init__()

        self.color = None
        self.shade = None
        self.opacity = None
        self.position = None

        self._quick_setup(kwargs)

    def __eq__(self, other):
        """
        Equality operator.
        """

        def dimcomp(a, b):
            if a is None:
                if b is not None:
                    return True
            else:
                if other.position is not None:
                    if a.value != b.value:
                        return True

            return False

        if self.color != other.color:
            return False

        if dimcomp(self.position, other.position):
            return False

        if dimcomp(self.shade, other.shade):
            return False

        if dimcomp(self.opacity, other.opacity):
            return False

        return True

    def _quick_setup(self, settings):
        """
        Method for defining gradient stop settings from class
        instanciation kwargs.

        :type settings: dict
        :param settings: Kwargs dictionnary
        """

        if settings:
            xmlc.PyScribusElement._quick_setup(self, settings)

            for setting_name, setting_value in settings.items():

                if setting_name == "color":
                    self.color = setting_value

                if setting_name == "position":
                    self.position = dimensions.Dim(
                        float(setting_value), unit="pcdecim"
                    )

                if setting_name == "opacity":
                    self.opacity = dimensions.Dim(
                        float(setting_value), unit="pcdecim"
                    )

                if setting_name == "shade":
                    self.shade = dimensions.Dim(float(setting_value), unit="pc")

    def fromdefault(self) -> NoReturn:
        self.opacity = dimensions.Dim(100, unit="pc")
        self.position = dimensions.Dim(0, unit="pcdecim")

    def fromxml(self, xml):
        if xml.tag == "CSTOP":

            color = xml.get("NAME")
            if color is not None:
                self.color = color

            shade = xml.get("SHADE")
            if shade is not None:
                self.shade = dimensions.Dim(float(shade), unit="pc")

            ramp = xml.get("RAMP")
            if ramp is not None:
                self.position = dimensions.Dim(float(ramp), unit="pcdecim")

            opacity = xml.get("TRANS")
            if opacity is not None:
                self.opacity = dimensions.Dim(float(opacity), "pcdecim")
                self.opacity = dimensions.Dim(float(opacity), unit="pcdecim")

            return True

        return False

    def toxml(self):
        xml = ET.Element("CSTOP")

        if self.position is not None:
            xml.attrib["RAMP"] = self.position.toxmlstr()

        xml.attrib["NAME"] = self.name

        if self.shade is None:
            xml.attrib["SHADE"] = "100"
        else:
            xml.attrib["SHADE"] = self.shade.toxmlstr(True)

        if self.opacity is None:
            xml.attrib["TRANS"] = "1"
        else:
            xml.attrib["TRANS"] = self.opacity.toxmlstr(True)

        return xml


class Gradient(xmlc.PyScribusElement):
    """
    Gradient in SLA (Gradient)
    """

    # <Gradient Name="Orange, Jaune" Ext="3">
    # </Gradient>

    def __init__(self):
        super().__init__()

        self.name = None
        self.stops = []

    def _sorted_stops(self):
        return sorted(self.stops, key=lambda s: s.position.value)

    def fromxml(self, xml):
        if xml.tag == "Gradient":

            name = xml.get("Name")
            if name is not None:
                self.name = name

            # TODO FIXME Ext

            for element in xml:
                if element.tag == "CSTOP":
                    grs = GradientColorStop()

                    success = grs.fromxml(element)
                    if success:
                        self.stops.append(grs)

            return True

        return False

    def append_stop(self, stop, sort=False):
        """
        Append a gradient color stop.

        Avoids duplicates. Can sort stops by color stop position.

        :type stop: pyscribus.colors.GradientColorStop
        :param stop: Gradient color stop to append
        :type sort: boolean
        :param sort: Sort gradient color stops by position.
        """

        if isinstance(stop, GradientColorStop):

            duplicate = [cs for cs in self.stops if cs == stop]

            if not duplicate:

                self.stops.append(stop)

                if sort:
                    self.stops = self._sorted_stops()

                return True

        return False

    def toxml(self):
        xml = ET.Element("Gradient")

        xml.attrib["Name"] = self.name

        # TODO FIXME Ext

        if self.stops:

            # Sort the color stops of the gradient by their position in
            # the gradient spectrum, ranging from 0 to 100%

            for stop in self._sorted_stops():
                sx = stop.toxml()
                xml.append(sx)

        return xml

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
ISO 216, international system for paper size.

A, B formats.

Yes, this is the module for A4 paper sizes.

See iso269paper module for C format.
See iso217paper module for RA, SRA formats.
"""

# Imports ===============================================================#

from pyscribus.common.math import FloatEnum, PICA_TO_MM

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

# Classes ===============================================================#

# A format --------------------------------------

class A0(FloatEnum):
    WIDTH = 841 / PICA_TO_MM
    HEIGHT = 1189 / PICA_TO_MM

class A1(FloatEnum):
    WIDTH = 594 / PICA_TO_MM
    HEIGHT = 841 / PICA_TO_MM

class A2(FloatEnum):
    WIDTH = 420 / PICA_TO_MM
    HEIGHT = 594 / PICA_TO_MM

class A3(FloatEnum):
    WIDTH = 297 / PICA_TO_MM
    HEIGHT = 420 / PICA_TO_MM

class A4(FloatEnum):
    WIDTH = 210 / PICA_TO_MM
    HEIGHT = 297 / PICA_TO_MM

class A5(FloatEnum):
    WIDTH = 148 / PICA_TO_MM
    HEIGHT = 210 / PICA_TO_MM

class A6(FloatEnum):
    WIDTH = 105 / PICA_TO_MM
    HEIGHT = 148 / PICA_TO_MM

class A7(FloatEnum):
    WIDTH = 74 / PICA_TO_MM
    HEIGHT = 105 / PICA_TO_MM

class A8(FloatEnum):
    WIDTH = 52 / PICA_TO_MM
    HEIGHT = 74 / PICA_TO_MM

class A9(FloatEnum):
    WIDTH = 37 / PICA_TO_MM
    HEIGHT = 52 / PICA_TO_MM

class A10(FloatEnum):
    WIDTH = 26 / PICA_TO_MM
    HEIGHT = 37 / PICA_TO_MM

# B format --------------------------------------

class B0(FloatEnum):
    WIDTH = 1000 / PICA_TO_MM
    HEIGHT = 1414 / PICA_TO_MM

class B1(FloatEnum):
    WIDTH = 707 / PICA_TO_MM
    HEIGHT = 1000 / PICA_TO_MM

class B2(FloatEnum):
    WIDTH = 500 / PICA_TO_MM
    HEIGHT = 707 / PICA_TO_MM

class B3(FloatEnum):
    WIDTH = 353 / PICA_TO_MM
    HEIGHT = 500 / PICA_TO_MM

class B4(FloatEnum):
    WIDTH = 250 / PICA_TO_MM
    HEIGHT = 353 / PICA_TO_MM

class B5(FloatEnum):
    WIDTH = 176 / PICA_TO_MM
    HEIGHT = 250 / PICA_TO_MM

class B6(FloatEnum):
    WIDTH = 125 / PICA_TO_MM
    HEIGHT = 176 / PICA_TO_MM

class B7(FloatEnum):
    WIDTH = 88 / PICA_TO_MM
    HEIGHT = 125 / PICA_TO_MM

class B8(FloatEnum):
    WIDTH = 62 / PICA_TO_MM
    HEIGHT = 88 / PICA_TO_MM

class B9(FloatEnum):
    WIDTH = 44 / PICA_TO_MM
    HEIGHT = 62 / PICA_TO_MM

class B10(FloatEnum):
    WIDTH = 31 / PICA_TO_MM
    HEIGHT = 44 / PICA_TO_MM

# vim:set shiftwidth=4 softtabstop=4 spl=en:



#!/usr/bin/python3
# -*- coding:Utf-8 -*-

# PyScribus, python library for Scribus SLA
# Copyright (C) 2020-2023 Étienne Nadji
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Mathematics related values, functions and enumerations.
"""

# Imports ===============================================================#

from typing import Union

import enum
import math

# Variables globales ====================================================#

__author__ = "Etienne Nadji <etnadji@eml.cc>"

PICA_TO_MM = (25.4 / 72)

INCH_TO_MM = 25.4

# Classes ===============================================================#

class FloatEnum(enum.Enum):
    """
    Enum class usable with float() function / method.
    """

    def __float__(self):
        if isinstance(self.value, float):
            return self.value
        else:
            return float(self.value)

# Fonctions =============================================================#


def truncate(number, digits) -> float:
    """
    Truncate value `number` to X `digits` digits.

    From Stack Overflow.

    <https://stackoverflow.com/questions/8595973/truncate-to-three-decimals-in-python>

    :type number: float
    :type digits: int
    :rtype: float
    """

    dec_number = len(str(number).split('.')[1])

    if dec_number <= digits:
        return number

    stepper = 10.0 ** digits

    return math.trunc(stepper * number) / stepper


def necessary_float(f: float) -> Union[float, int]:
    """
    Return integer if float f has no decimals.

    :type f: float
    :param f: Float value
    :rtype: int,float
    :returns: Integer if float f has no decimals, else float.
    """

    if float(f) == int(f):
        return int(f)
    else:
        return float(f)


def mm(milimeters):
    """
    Returns milimeters in pica points.
    """

    return float(milimeters) / PICA_TO_MM

# vim:set shiftwidth=4 softtabstop=4:
